"""Test functionalities of flux analysis helper functions."""

import pytest

from cobra.core import Model
from cobra.flux_analysis.helpers import normalize_cutoff


def test_normalize_cutoff(model: Model) -> None:
    """Test normalize cutoff."""
    cutoff = normalize_cutoff(model)
    assert cutoff == 1e-7


def test_normalize_cutoff_with_specified_cutoff_above_default(
    model: Model,
) -> None:
    """Test normalize cutoff with specified cutoff greater than default."""
    cutoff = normalize_cutoff(model, 1e-3)
    assert cutoff == 1e-3


def test_normalize_cutoff_with_specified_cutoff_below_default(
    model: Model,
) -> None:
    """Test normalize cutoff with specified cutoff less than default."""
    with pytest.raises(ValueError):
        normalize_cutoff(model, 1e-10)



"""Unit test the MetaboliteSummary class."""

import pytest

from cobra.flux_analysis import flux_variability_analysis, pfba
from cobra.summary import MetaboliteSummary


def test_metabolite_summary_interface(model, opt_solver):
    """Test that a summary can be created successfully."""
    model.solver = opt_solver
    metabolite = model.metabolites.get_by_id("q8_c")
    MetaboliteSummary(
        metabolite=metabolite,
        model=model,
    )
    MetaboliteSummary(
        metabolite=metabolite,
        model=model,
        solution=pfba(model),
    )
    MetaboliteSummary(
        metabolite=metabolite,
        model=model,
        fva=0.95,
    )
    MetaboliteSummary(
        metabolite=metabolite,
        model=model,
        fva=flux_variability_analysis(
            model, reaction_list=["CYTBD", "NADH16", "SUCDi"]
        ),
    )


def test_metabolite_summary_to_frame(model, opt_solver):
    """Test that the summary's method ``to_frame`` can be called."""
    model.solver = opt_solver
    summary = model.metabolites.get_by_id("atp_c").summary()
    summary.to_frame()


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"names": True},
        {"float_format": ".1f"},
        {"threshold": 1e-2},
        {"column_width": 20},
    ],
)
def test_metabolite_summary_to_string(model, opt_solver, kwargs):
    """Test that the summary's method ``to_string`` can be called."""
    model.solver = opt_solver
    summary = model.metabolites.get_by_id("atp_c").summary()
    summary.to_string(**kwargs)


@pytest.mark.parametrize(
    "kwargs", [{}, {"names": True}, {"float_format": ".1f"}, {"threshold": 1e-2}]
)
def test_metabolite_summary_to_html(model, opt_solver, kwargs):
    """Test that the summary's method ``to_html`` can be called."""
    model.solver = opt_solver
    summary = model.metabolites.get_by_id("atp_c").summary()
    summary.to_html(**kwargs)


def test_q8_producing_summary(model, opt_solver):
    """Test that the production summary of q8 is accurate."""
    model.solver = opt_solver
    summary = model.metabolites.get_by_id("q8_c").summary()

    assert summary.producing_flux.at["CYTBD", "percent"] == 1
    assert summary.producing_flux.at["CYTBD", "flux"] == pytest.approx(43.6, abs=1e-2)


def test_q8_consuming_summary(model, opt_solver):
    """Test that the consumption summary of q8 is accurate."""
    model.solver = opt_solver
    summary = model.metabolites.get_by_id("q8_c").summary()

    assert summary.consuming_flux.at["NADH16", "percent"] == pytest.approx(
        0.8838, abs=1e-4
    )
    assert summary.consuming_flux.at["NADH16", "flux"] == pytest.approx(
        -38.54, abs=1e-2
    )

    assert summary.consuming_flux.at["SUCDi", "percent"] == pytest.approx(
        0.1162, abs=1e-4
    )
    assert summary.consuming_flux.at["SUCDi", "flux"] == pytest.approx(-5.06, abs=1e-2)


def test_fdp_production_with_fva(model, opt_solver):
    """Test that the production summary of fdp is within expected bounds."""
    model.solver = opt_solver
    summary = model.metabolites.get_by_id("fdp_c").summary(fva=0.99)
    assert summary.producing_flux.at["PFK", "percent"] == 1
    assert summary.producing_flux.at["PFK", "flux"] == pytest.approx(7.48, abs=1e-2)
    assert summary.producing_flux.at["PFK", "minimum"] == pytest.approx(6.17, abs=1e-2)
    assert summary.producing_flux.at["PFK", "maximum"] == pytest.approx(9.26, abs=1e-2)


@pytest.mark.parametrize("metabolite_id", ["q8_c", "fdp_c", "atp_c"])
def test_metabolite_summary_flux_in_context(model, opt_solver, metabolite_id: str):
    """Test that the metabolite summary inside and outside of a context are equal."""
    model.solver = opt_solver
    with model:
        context_summary = model.metabolites.get_by_id(metabolite_id).summary()
    outside_summary = model.metabolites.get_by_id(metabolite_id).summary()

    assert context_summary.to_frame()["flux"].values == pytest.approx(
        outside_summary.to_frame()["flux"].values, abs=model.tolerance
    )



"""Define module level fixtures."""

from typing import TYPE_CHECKING, Tuple

import pytest

from cobra.util.solver import solvers


if TYPE_CHECKING:
    from cobra import Model, Solution


solver_trials = [
    "glpk",
    pytest.param(
        "cplex",
        marks=pytest.mark.skipif(
            "cplex" not in solvers, reason="No CPLEX found on PYTHONPATH"
        ),
    ),
    pytest.param(
        "gurobi",
        marks=pytest.mark.skipif(
            "gurobi" not in solvers, reason="No Gurobi found on PYTHONPATH"
        ),
    ),
]


@pytest.fixture(scope="function", params=solver_trials)
def solved_model(
    request: pytest.FixtureRequest, model: "Model"
) -> Tuple["Solution", "Model"]:
    """Return solved model."""
    model.solver = request.param
    solution = model.optimize()
    return solution, model



"""Test functions of cobra.core.gene.GPR ."""

import itertools
from ast import parse as ast_parse
from typing import Dict, Iterable, Iterator, List, Set, Tuple, Union

import pytest
from sympy.core.expr import Expr
from sympy.core.symbol import Symbol
from sympy.logic import And, Or
from sympy.logic.boolalg import BooleanFunction

from cobra.core.gene import GPR, ast2str, eval_gpr, parse_gpr
from cobra.core.model import Model


def test_gpr() -> None:
    """Test GPR instance creation and basic usage."""
    gpr1 = GPR()
    assert len(gpr1.genes) == 0
    gpr1.update_genes()
    assert len(gpr1.genes) == 0
    assert gpr1.to_string() == ""
    assert gpr1.eval()
    gpr2 = gpr1.copy()
    assert isinstance(gpr2, GPR)
    assert len(gpr1.genes) == 0


def test_grp_copy() -> None:
    """Test that copying a GPR works."""
    gpr1 = GPR.from_string("(A and B) or C")
    gpr2 = gpr1.copy()
    assert gpr1 == gpr2
    assert id(gpr1.body) is not id(gpr2.body)
    assert gpr1._genes == gpr2._genes
    assert id(gpr1._genes) is not id(gpr2._genes)


@pytest.mark.parametrize("test_input", ["", "", None])
def test_empty_gpr(test_input) -> None:
    """Test empty GPR."""
    gpr1 = GPR(test_input)
    assert not gpr1.body
    assert len(gpr1.genes) == 0
    gpr1.update_genes()
    assert len(gpr1.genes) == 0
    assert gpr1.to_string() == ""
    assert gpr1.eval()
    assert len(gpr1.genes) == 0


def test_one_gene_gpr() -> None:
    """Test single gene GPR."""
    gpr1 = GPR.from_string("a")
    assert len(gpr1.genes) == 1
    gpr1.update_genes()
    assert len(gpr1.genes) == 1
    assert gpr1.genes == {"a"}
    assert gpr1.to_string() == "a"
    assert gpr1.eval()
    assert len(gpr1.genes) == 1


def powerset_ne(iterable: Iterable[str]) -> Iterator[Tuple[str, ...]]:
    """Get all combinations of an iterable except the empty list.

    Gets an iterable of all combinations of genes except the empty list.
    Used to evaluate AND gprs
    powerset_ne([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(1, len(s) + 1)
    )


@pytest.mark.parametrize(
    "gpr_input, num_genes, gpr_genes, gpr_output_string",
    [
        ("a & b", 2, {"a", "b"}, "a and b"),
        ("a and b", 2, {"a", "b"}, "a and b"),
        pytest.param(
            "a AND b", 2, {"a", "b"}, "a and b", marks=pytest.mark.filterwarnings
        ),
        ("a & b & c", 3, {"a", "b", "c"}, "(a and b) and c"),
        ("a and b and c", 3, {"a", "b", "c"}, "a and b and c"),
    ],
)
def test_and_gpr(gpr_input, num_genes, gpr_genes, gpr_output_string) -> None:
    """Test 'and' GPR."""
    gpr1 = GPR.from_string(gpr_input)
    assert len(gpr1.genes) == num_genes
    gpr1.update_genes()
    assert len(gpr1.genes) == num_genes
    assert gpr1.genes == gpr_genes
    assert gpr1.to_string() == gpr_output_string
    assert gpr1.eval()
    for ko_genes in powerset_ne(gpr_genes):
        assert not gpr1.eval(ko_genes)
    assert gpr1.body


def all_except_one(iterable: Iterable[str]) -> Iterator[Tuple[str, ...]]:
    """Generate all combinations from an iterable, while leaving one out.

    Gets an iterable of all combinations of genes except the complete list and the
    empty list. Used to evaluate OR gprs
    all_except_one([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3)
    """
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(1, len(s))
    )


@pytest.mark.parametrize(
    "gpr_input, num_genes, gpr_genes, gpr_output_string",
    [
        ("a | b", 2, {"a", "b"}, "a or b"),
        ("a or b", 2, {"a", "b"}, "a or b"),
        pytest.param(
            "a OR b", 2, {"a", "b"}, "a or b", marks=pytest.mark.filterwarnings
        ),
        ("a | b | c", 3, {"a", "b", "c"}, "(a or b) or c"),
        ("a or b or c", 3, {"a", "b", "c"}, "a or b or c"),
    ],
)
def test_or_gpr(
    gpr_input: str, num_genes: int, gpr_genes: Set, gpr_output_string: str
) -> None:
    """Test 'or' GPR."""
    gpr1 = GPR.from_string(gpr_input)
    assert len(gpr1.genes) == num_genes
    gpr1.update_genes()
    assert len(gpr1.genes) == num_genes
    assert gpr1.genes == gpr_genes
    assert gpr1.to_string() == gpr_output_string
    assert gpr1.eval()
    for ko_genes in all_except_one(gpr_genes):
        assert gpr1.eval(ko_genes)
    assert not gpr1.eval(gpr_genes)
    assert gpr1.body


@pytest.mark.parametrize(
    "gpr_input",
    [
        "(a | b) & c",
        "(a or b) and c",
        pytest.param("(a OR b) AND c", marks=pytest.mark.filterwarnings),
    ],
)
def test_complicated_gpr(gpr_input: str) -> None:
    """Test complicated GPR."""
    gpr1 = GPR.from_string(gpr_input)
    assert len(gpr1.genes) == 3
    gpr1.update_genes()
    assert len(gpr1.genes) == 3
    assert gpr1.to_string() == "(a or b) and c"
    assert gpr1.eval()
    assert gpr1.body
    assert gpr1.eval()
    assert gpr1.eval("a")
    assert gpr1.eval("b")
    assert not gpr1.eval("c")
    assert not gpr1.eval(["a", "b"])
    assert not gpr1.eval(["a", "b", "c"])


@pytest.mark.parametrize(
    "string_to_ast, num_genes, gpr_genes, gpr_output_string",
    [
        ("a", 1, {"a"}, "a"),
        ("a | b", 2, {"a", "b"}, "a or b"),
        ("a or b", 2, {"a", "b"}, "a or b"),
        pytest.param("a OR b", 2, {"a", "b"}, "a or b", marks=pytest.mark.xfail),
    ],
)
def test_gpr_from_ast_or(
    string_to_ast: str, num_genes: int, gpr_genes: set, gpr_output_string: str
) -> None:
    """Test GPR from AST 'or'."""
    ast_tree = ast_parse(string_to_ast, "<string>", "eval")
    gpr1 = GPR(ast_tree)
    assert len(gpr1.genes) == num_genes
    gpr1.update_genes()
    assert len(gpr1.genes) == num_genes
    assert gpr1.genes == gpr_genes
    assert gpr1.to_string() == gpr_output_string
    assert gpr1.eval()
    for ko_genes in all_except_one(gpr_genes):
        assert gpr1.eval(ko_genes)
    assert not gpr1.eval(gpr_genes)


@pytest.mark.parametrize(
    "string_to_ast, num_genes, gpr_genes, gpr_output_string",
    [
        ("a & b", 2, {"a", "b"}, "a and b"),
        ("a and b", 2, {"a", "b"}, "a and b"),
        pytest.param("a AND b", 2, {"a", "b"}, "a and b", marks=pytest.mark.xfail),
    ],
)
def test_gpr_from_ast_and(
    string_to_ast: str, num_genes: int, gpr_genes: set, gpr_output_string: str
) -> None:
    """Test GPR from AST 'and'."""
    ast_tree = ast_parse(string_to_ast, "<string>", "eval")
    gpr1 = GPR(ast_tree)
    assert len(gpr1.genes) == num_genes
    gpr1.update_genes()
    assert len(gpr1.genes) == num_genes
    assert gpr1.genes == gpr_genes
    assert gpr1.to_string() == gpr_output_string
    assert gpr1.eval()
    for ko_genes in powerset_ne(gpr_genes):
        assert not gpr1.eval(ko_genes)


@pytest.mark.parametrize("test_input", [["a", "b"], {"a", "b"}])
def test_wrong_input_gpr_error(test_input: Union[list, set]) -> None:
    """Test error for incorrect GPR input."""
    with pytest.raises(TypeError):
        GPR.from_string(test_input)
    with pytest.raises(TypeError):
        GPR(test_input)


@pytest.mark.parametrize("test_input", ["a |", "a &", "a and ()", "a or ()"])
def test_wrong_input_gpr_warning(test_input: str) -> None:
    """Test warning for incorrect GPR input."""
    with pytest.warns(SyntaxWarning):
        gpr1 = GPR.from_string(test_input)
        assert gpr1.body is None
        assert len(gpr1.genes) == 0


def test_gpr_that_needs_two_replacements() -> None:
    """Test GPR with multi replacements."""
    gpr1 = GPR.from_string(
        "(591001.3.peg.1891 AND 591001.3.peg.1892 AND 591001.3.peg.1893)"
    )
    assert len(gpr1.genes) == 3
    assert "591001.3.peg.1891" in gpr1.genes
    assert "591001.3.peg.1892" in gpr1.genes
    assert "591001.3.peg.1893" in gpr1.genes


def test_deprecated_gpr() -> None:
    """Test deprecated GPR."""
    gpr1 = GPR.from_string("(a | b) & c")
    with pytest.deprecated_call():
        assert ast2str(gpr1) == "(a or b) and c"
    with pytest.deprecated_call():
        assert eval_gpr(gpr1, set())
    with pytest.deprecated_call():
        assert eval_gpr(gpr1, {"a"})
    with pytest.deprecated_call():
        assert eval_gpr(gpr1, {"b"})
    with pytest.deprecated_call():
        assert not eval_gpr(gpr1, {"c"})
    with pytest.deprecated_call():
        gpr1, genes = parse_gpr("(a | b) & c")
        assert genes == {"a", "b", "c"}
    with pytest.deprecated_call():
        assert ast2str(gpr1) == "(a or b) and c"
    with pytest.deprecated_call():
        assert eval_gpr(gpr1, set())
    with pytest.deprecated_call():
        assert eval_gpr(gpr1, {"a"})
    with pytest.deprecated_call():
        assert eval_gpr(gpr1, {"b"})
    with pytest.deprecated_call():
        assert not eval_gpr(gpr1, {"c"})


def test_gpr_as_symbolic() -> None:
    """Test GPR as symbolic expression."""
    gpr1 = GPR()
    assert gpr1.as_symbolic() == Symbol("")
    gpr1 = GPR.from_string("")
    assert gpr1.as_symbolic() == Symbol("")
    gpr1 = GPR.from_string("a")
    assert isinstance(gpr1.as_symbolic(), Symbol)
    assert gpr1.as_symbolic() == Symbol("a")
    gpr2 = GPR.from_string("a & b")
    assert isinstance(gpr2.as_symbolic(), BooleanFunction)
    assert gpr2.as_symbolic() == And(Symbol("a"), Symbol("b"))
    assert gpr1 != gpr2


@pytest.mark.parametrize(
    "gpr_input, symbolic_gpr",
    [
        ("a | b", Or(Symbol("a"), Symbol("b"))),
        ("a or b", Or(Symbol("a"), Symbol("b"))),
        pytest.param(
            "a OR b", Or(Symbol("a"), Symbol("b")), marks=pytest.mark.filterwarnings
        ),
        ("a | b", Or(Symbol("b"), Symbol("a"))),
        ("a | b | c", Or(Symbol("a"), Symbol("b"), Symbol("c"))),
        ("a or b or c", Or(Symbol("a"), Symbol("b"), Symbol("c"))),
        ("(a OR b) AND c", And(Symbol("c"), Or(Symbol("a"), Symbol("b")))),
    ],
)
def test_gpr_as_symbolic_boolean(gpr_input: str, symbolic_gpr: Expr) -> None:
    """Test GPR as symbolic boolean expression."""
    gpr1 = GPR().from_string(gpr_input)
    assert isinstance(gpr1.as_symbolic(), BooleanFunction)
    assert gpr1.as_symbolic() == symbolic_gpr
    if "OR" not in gpr_input and "AND" not in gpr_input:
        ast_tree = ast_parse(gpr_input, "<string>", "eval")
        gpr1 = GPR(ast_tree)
        assert isinstance(gpr1.as_symbolic(), BooleanFunction)
        assert gpr1.as_symbolic() == symbolic_gpr


def test_gpr_equality() -> None:
    """Test GPR equality."""
    assert GPR() == GPR()
    assert GPR() == GPR.from_string("")
    assert GPR() != GPR.from_string("a")
    assert GPR.from_string("a") == GPR.from_string("a")


gpr_str_lists = {
    "a_and_b_strs": ["a & b", "a and b", "a AND b", "b & a", "b and a", "b AND a"],
    "a_or_b_strs": ["a | b", "a or b", "a OR b", "b | a", "b or a", "b OR a"],
    "a_b_c_or_strs": [
        "a | b | c",
        "a or b or c",
        "a OR b or c",
        "(a or b) or c",
        "b | a | c",
        "b or a or c",
        "b | c | a",
        "b or c | a",
        "b or c or a",
        "c or a or b",
        "c | a or b",
        "c or a | b",
        "c or b or a",
        "c | b or a",
        "c or b | a",
        "c | b | a",
    ],
    "a_b_c_and_strs": [
        "a & b & c",
        "a and b and c",
        "a AND b and c",
        "(a and b) and c",
        "b & a & c",
        "b and a and c",
        "b & c & a",
        "b and c & a",
        "b and c and a",
        "c and a and b",
        "c & a and b",
        "c and a & b",
        "c and b and a",
        "c & b and a",
        "c and b & a",
        "c & b & a",
    ],
    "a_b_c_or_and_strs": [
        "(a OR b) AND c",
        "(a | b) & c",
        "(a & c) | b & c",
        "c & (a | b)",
    ],
}


@pytest.fixture(params=list(gpr_str_lists.keys()))
def gpr_list(request: pytest.FixtureRequest) -> List[str]:
    """Provide fixture for GPR list."""
    gpr_str_list = gpr_str_lists[request.param]
    return gpr_str_list


def test_gpr_equality_with_bolean_logic(gpr_list: List[str]) -> None:
    """Test GPR equality with boolean logic."""
    for i in range(len(gpr_list)):
        for j in range(i + 1, len(gpr_list)):
            assert GPR().from_string(gpr_list[i]) == GPR.from_string(gpr_list[j])


@pytest.fixture(params=list(itertools.combinations(gpr_str_lists.keys(), 2)))
def gpr_lists(request: pytest.FixtureRequest) -> Dict[str, List[str]]:
    """Provide fixture for GPR dictionary."""
    gpr_dict = dict()
    gpr_dict["gpr1"] = gpr_str_lists[request.param[0]]
    gpr_dict["gpr2"] = gpr_str_lists[request.param[1]]
    return gpr_dict


def test_gpr_inequality_boolean(gpr_lists) -> None:
    """Test GPR inequality for boolean expression."""
    gpr_list1 = gpr_lists["gpr1"]
    gpr_list2 = gpr_lists["gpr2"]
    for i in range(len(gpr_list1)):
        for j in range(len(gpr_list2)):
            assert GPR.from_string(gpr_list1[i]) != GPR.from_string(gpr_list2[j])


def test_gpr_symbolism_benchmark(large_model: Model, benchmark) -> None:
    """Benchmark as symbolic time."""
    model = large_model.copy()

    def gpr_symbolic():
        for i in range(len(model.reactions)):
            rxn1 = model.reactions[i]
            gpr1 = rxn1.gpr
            gpr1.as_symbolic()

    benchmark(gpr_symbolic)


def test_gpr_equality_benchmark(model: Model, benchmark) -> None:
    """Benchmark equality of GPR using the mini model."""
    model2 = model.copy()

    def gpr_equality_all_reactions():
        for i in range(len(model.reactions)):
            rxn1 = model.reactions[i]
            rxn2 = model2.reactions[i]
            assert rxn1.gpr == rxn2.gpr

    benchmark(gpr_equality_all_reactions)


@pytest.mark.parametrize(
    "gpr_input, symbolic_gpr",
    [
        ("", Symbol("")),
        ("a", Symbol("a")),
        ("a & b", And(Symbol("a"), Symbol("b"))),
        ("a | b", Or(Symbol("a"), Symbol("b"))),
        ("a or b", Or(Symbol("a"), Symbol("b"))),
        pytest.param(
            "a OR b", Or(Symbol("a"), Symbol("b")), marks=pytest.mark.filterwarnings
        ),
        ("a | b", Or(Symbol("b"), Symbol("a"))),
        ("a | b | c", Or(Symbol("a"), Symbol("b"), Symbol("c"))),
        ("a or b or c", Or(Symbol("a"), Symbol("b"), Symbol("c"))),
        ("(a OR b) AND c", And(Symbol("c"), Or(Symbol("a"), Symbol("b")))),
    ],
)
def test_gpr_from_symbolic(gpr_input: str, symbolic_gpr: Expr) -> None:
    """Test GPR creation from symbolic and from string is equal."""
    gpr1 = GPR().from_symbolic(symbolic_gpr)
    gpr2 = GPR().from_string(gpr_input)
    assert gpr1 == gpr2


def test_gpr_from_as_symbolic_equality(large_model) -> None:
    """Test as_symbolic followed by from_symbolic gives a GPR equivalent to original."""
    model = large_model.copy()

    for i in range(len(model.reactions)):
        rxn1 = model.reactions[i]
        gpr1 = rxn1.gpr
        gpr2 = GPR().from_symbolic(gpr1.as_symbolic())
        assert gpr1 == gpr2



"""Test functions of cobra.core.reaction ."""

import warnings
from typing import Iterable

import numpy as np
import pytest

from cobra import Gene
from cobra.core import GPR, Configuration, Metabolite, Model, Reaction


config = Configuration()
stable_optlang = ["glpk", "cplex", "gurobi"]


def test_gpr() -> None:
    """Test GPR evaluation."""
    model = Model()
    reaction = Reaction("test")

    # Set GPR to an empty string
    reaction.gene_reaction_rule = ""
    # Empty gene_reaction_rule leads to an empty GPR
    assert reaction.gpr.body is None
    assert reaction.gpr.to_string() == ""
    assert reaction.gpr.to_string(names={"goo": "blah"}) == ""
    # Set GPR directly (shouldn't really be used, but just a test)
    reaction.gpr = GPR()
    assert reaction.gene_reaction_rule == ""
    assert reaction.gpr.eval()
    # Set GPR to a reaction not in a model
    reaction.gene_reaction_rule = "(g1 or g2) and g3"
    assert reaction.gene_reaction_rule == "(g1 or g2) and g3"
    assert len(reaction.genes) == 3

    # Adding reaction with a GPR propagates to the model
    model.add_reactions([reaction])
    assert len(model.genes) == 3

    # Ensure the gene objects are the same in the model and reaction
    reaction_gene = list(reaction.genes)[0]
    model_gene = model.genes.get_by_id(reaction_gene.id)
    assert reaction_gene is model_gene


def test_gpr_uppercase() -> None:
    """Test ability to handle uppercase AND/OR."""
    reaction = Reaction("test")
    with pytest.warns(SyntaxWarning):
        reaction.gene_reaction_rule = "(b1 AND b2) OR (b3 and b4)"
        assert reaction.gene_reaction_rule == "(b1 and b2) or (b3 and b4)"
        assert len(reaction.genes) == 4


@pytest.mark.parametrize("input_gpr", ["(a1 or a2", "(forT or "])
def test_gpr_malformed(input_gpr: str) -> None:
    """Test ability to deal with malformed GPR.

    Malformed GPR strings will lead to empty GPRs with no genes.

    Parameters
    ----------
    input_gpr: str
        String representing a malformed GPR string.
    """
    reaction = Reaction("test")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        reaction.gene_reaction_rule = input_gpr
        assert len(reaction.genes) == 0


def test_gpr_modification(model: Model) -> None:
    """Test GPR manipulations."""
    reaction = model.reactions.get_by_id("PGI")
    old_gene = list(reaction.genes)[0]
    new_gene = model.genes.get_by_id("s0001")

    # Add an existing 'gene' to the GPR
    reaction.gene_reaction_rule = "s0001"
    assert new_gene in reaction.genes
    assert reaction in new_gene.reactions

    # Remove old gene correctly, keep it in model
    assert old_gene not in reaction.genes
    assert reaction not in old_gene.reactions
    assert old_gene in model.genes

    # Add a new 'gene' to the GPR
    reaction.gene_reaction_rule = "fake_gene"
    assert model.genes.has_id("fake_gene")
    fake_gene = model.genes.get_by_id("fake_gene")
    assert fake_gene in reaction.genes
    assert reaction in fake_gene.reactions
    fake_gene.name = "foo_gene"
    assert reaction.gene_name_reaction_rule == fake_gene.name


def test_gpr_modification_with_context(model: Model) -> None:
    """Test GPR manipulations are reversed in context."""
    empty_model = Model()
    reaction = Reaction("test")
    reaction.gene_reaction_rule = "(g1 or g2) and g3"
    assert reaction.gene_reaction_rule == "(g1 or g2) and g3"
    assert len(reaction.genes) == 3

    with empty_model:
        # Adding reaction with a GPR propagates to the model
        empty_model.add_reactions([reaction])
        assert len(empty_model.genes) == 3
    assert len(empty_model.reactions) == 0
    assert len(empty_model.genes) == 0
    assert reaction._model is None

    reaction = model.reactions.get_by_id("PGI")
    old_reaction_rule = reaction.gene_reaction_rule
    old_gene = list(reaction.genes)[0]
    new_gene = model.genes.get_by_id("s0001")

    with model:
        # Add an existing 'gene' to the GPR
        reaction.gene_reaction_rule = "s0001"
        assert new_gene in reaction.genes
        assert reaction in new_gene.reactions

        # Remove old gene correctly, keep it in model
        assert old_gene not in reaction.genes
        assert reaction not in old_gene.reactions
        assert old_gene in model.genes

    assert reaction.gene_reaction_rule == old_reaction_rule
    assert new_gene not in reaction.genes
    assert reaction not in new_gene.reactions

    # Remove old gene correctly, keep it in model
    assert old_gene in reaction.genes
    assert reaction in old_gene.reactions
    assert old_gene in model.genes

    with model:
        # Add a new 'gene' to the GPR
        reaction.gene_reaction_rule = "fake_gene"
        assert model.genes.has_id("fake_gene")
        fake_gene = model.genes.get_by_id("fake_gene")
        assert fake_gene in reaction.genes
        assert reaction in fake_gene.reactions
        fake_gene.name = "foo_gene"
        assert reaction.gene_name_reaction_rule == fake_gene.name
    assert not model.genes.has_id("fake_gene")
    fake_gene = Gene("fake_gene")
    assert fake_gene not in reaction.genes
    assert reaction not in fake_gene.reactions


def test_gene_knock_out(model: Model) -> None:
    """Test gene knockout effect on reaction."""
    rxn = Reaction("rxn")
    rxn.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    rxn.gene_reaction_rule = "A2B1 or A2B2 and A2B3"
    assert hasattr(list(rxn.genes)[0], "knock_out")
    model.add_reactions([rxn])
    with model:
        model.genes.A2B1.knock_out()
        assert not model.genes.A2B1.functional
        model.genes.A2B3.knock_out()
        assert not rxn.functional
    assert model.genes.A2B3.functional
    assert rxn.functional
    model.genes.A2B1.knock_out()
    assert not model.genes.A2B1.functional
    assert model.reactions.rxn.functional
    model.genes.A2B3.knock_out()
    assert not model.reactions.rxn.functional


def test_str() -> None:
    """Test `str` output for a reaction."""
    rxn = Reaction("rxn")
    rxn.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    assert str(rxn) == "rxn: A --> B"


def test_str_from_model(model: Model) -> None:
    """Test `str` output for a reaction associated with a model."""
    assert model.reactions[0].__str__().startswith("ACALD")


def test_add_metabolite_from_solved_model(solved_model: Model) -> None:
    """Test metabolite addition to a reaction from a solved model."""
    solution, model = solved_model
    pgi_reaction = model.reactions.PGI
    test_met = model.metabolites[0]
    pgi_reaction.add_metabolites({test_met: 42}, combine=False)
    assert pgi_reaction.metabolites[test_met] == 42.0
    assert (
        model.constraints[test_met.id].expression.as_coefficients_dict()[
            pgi_reaction.forward_variable
        ]
        == 42.0
    )
    assert (
        model.constraints[test_met.id].expression.as_coefficients_dict()[
            pgi_reaction.reverse_variable
        ]
        == -42.0
    )

    pgi_reaction.add_metabolites({test_met: -10}, combine=True)
    assert pgi_reaction.metabolites[test_met] == 32.0
    assert (
        model.constraints[test_met.id].expression.as_coefficients_dict()[
            pgi_reaction.forward_variable
        ]
        == 32.0
    )
    assert (
        model.constraints[test_met.id].expression.as_coefficients_dict()[
            pgi_reaction.reverse_variable
        ]
        == -32.0
    )

    pgi_reaction.add_metabolites({test_met: 0}, combine=False)
    with pytest.raises(KeyError):
        pgi_reaction.metabolites[test_met]
    assert (
        model.constraints[test_met.id].expression.as_coefficients_dict()[
            pgi_reaction.forward_variable
        ]
        == 0
    )
    assert (
        model.constraints[test_met.id].expression.as_coefficients_dict()[
            pgi_reaction.reverse_variable
        ]
        == 0
    )


@pytest.mark.parametrize("solver", stable_optlang)
def test_add_metabolite_benchmark(model: Model, benchmark, solver: Iterable) -> None:
    """Benchmark metabolite addition to a reaction associated with a model."""
    reaction = model.reactions.get_by_id("PGI")
    many_metabolites = dict((m, 1) for m in model.metabolites[0:50])

    def add_remove_metabolite():
        reaction.add_metabolites(many_metabolites)
        if not getattr(model, "solver", None):
            stable_optlang[solver].create_problem(model)
        for met in many_metabolites:
            try:
                reaction.subtract_metabolites({met: reaction.get_coefficient(met)})
            except KeyError:
                pass

    benchmark(add_remove_metabolite)


def test_add_metabolite(model: Model) -> None:
    """Test metabolite addition to a reaction from an unsolved model."""
    with pytest.raises(ValueError):
        model.add_metabolites(Metabolite())
    with model:
        with model:
            reaction = model.reactions.get_by_id("PGI")
            reaction.add_metabolites({model.metabolites[0]: 1})
            assert model.metabolites[0] in reaction._metabolites
            fake_metabolite = Metabolite("fake")
            reaction.add_metabolites({fake_metabolite: 1})
            assert fake_metabolite in reaction._metabolites
            assert model.metabolites.has_id("fake")
            assert model.metabolites.get_by_id("fake") is fake_metabolite
            assert len(model._contexts[0]._history) == 0

    assert fake_metabolite._model is None
    assert fake_metabolite not in reaction._metabolites
    assert "fake" not in model.metabolites

    # Test adding by string
    with model:
        reaction.add_metabolites({"g6p_c": -1})  # already in reaction
        assert reaction._metabolites[model.metabolites.get_by_id("g6p_c")] == -2
        reaction.add_metabolites({"h_c": 1})
        assert reaction._metabolites[model.metabolites.get_by_id("h_c")] == 1
        with pytest.raises(KeyError):
            reaction.add_metabolites({"missing": 1})

    assert reaction._metabolites[model.metabolites.get_by_id("g6p_c")] == -1
    assert model.metabolites.h_c not in reaction._metabolites

    # Test combine=False
    reaction = model.reactions.get_by_id("ATPM")
    old_stoich = reaction._metabolites[model.metabolites.get_by_id("h2o_c")]
    with model:
        reaction.add_metabolites({"h2o_c": 2.5}, combine=False)
        assert reaction._metabolites[model.metabolites.get_by_id("h2o_c")] == 2.5

    assert reaction._metabolites[model.metabolites.get_by_id("h2o_c")] == old_stoich

    # Test adding to a new Reaction
    reaction = Reaction("test")
    assert len(reaction._metabolites) == 0
    reaction.add_metabolites({Metabolite("test_met"): -1})
    assert len(reaction._metabolites) == 1


@pytest.mark.parametrize("solver", stable_optlang)
def test_subtract_metabolite_benchmark(
    model: Model, benchmark, solver: Iterable
) -> None:
    """Benchmark metabolite deletion from a reaction."""
    benchmark(test_subtract_metabolite, model, solver)


@pytest.mark.parametrize("solver", stable_optlang)
def test_subtract_metabolite(model: Model, solver: Iterable) -> None:
    """Test metabolite deletion from a reaction associated with an unsolved model."""
    reaction = model.reactions.get_by_id("PGI")
    reaction.subtract_metabolites(reaction.metabolites)
    if not getattr(model, "solver", None):
        stable_optlang[solver].create_problem(model)
        assert len(reaction.metabolites) == 0


def test_mass_balance(model: Model) -> None:
    """Test mass balance of metabolites of a reaction."""
    reaction = model.reactions.get_by_id("PGI")
    # Should be balanced now
    assert len(reaction.check_mass_balance()) == 0
    # Should not be balanced after adding a hydrogen
    reaction.add_metabolites({model.metabolites.get_by_id("h_c"): 1})
    imbalance = reaction.check_mass_balance()
    assert "H" in imbalance
    assert imbalance["H"] == 1
    # Should be balanced even when coefficients are non-integer
    reaction_scaled = model.reactions.get_by_id("ATPS4r") * 1.3
    assert len(reaction_scaled.check_mass_balance()) == 0


def test_build_from_string(model: Model) -> None:
    """Test reaction building from string evaluation."""
    m = len(model.metabolites)
    pgi = model.reactions.get_by_id("PGI")
    old_bounds = pgi.bounds

    with model:
        pgi.reaction = "g6p_c --> f6p_c"
        assert pgi.lower_bound == 0

    assert pgi.bounds == old_bounds

    pgi.bounds = (0, 1000)
    assert pgi.bounds == (0, 1000)
    assert not pgi.reversibility
    pgi.reaction = "g6p_c <== f6p_c"
    assert pgi.upper_bound == 0
    assert pgi.reaction.strip() == "g6p_c <-- f6p_c"
    pgi.reaction = "g6p_c --> f6p_c + h2o_c"
    assert model.metabolites.h2o_c, pgi._metabolites

    with model:
        pgi.build_reaction_from_string("g6p_c --> f6p_c + foo", verbose=False)
        assert model.metabolites.h2o_c not in pgi._metabolites
        assert "foo" in model.metabolites
        assert model.metabolites.foo in pgi._metabolites
        assert len(model.metabolites) == m + 1

    assert model.metabolites.h2o_c in pgi._metabolites
    assert "foo" not in model.metabolites
    with pytest.raises(AttributeError):
        assert model.metabolites.foo
        assert len(model.metabolites) == m

    with model:
        old_bounds = config.bounds
        assert old_bounds == (-1000, 1000)
        config.bounds = (-5, 5)
        pgi.build_reaction_from_string("g6p_c <--> f6p_c + new", verbose=False)
        assert pgi.bounds == (-5, 5)
        config.bounds = old_bounds
        pgi.build_reaction_from_string("g6p_c --> f6p_c + new", verbose=False)
        assert pgi.bounds == (0, 1000)


def test_build_from_string_creating_metabolites() -> None:
    """Test that metabolites are created in the correct compartment."""
    # https://github.com/opencobra/cobrapy/issues/1418
    model = Model()
    reaction = Reaction("R1")
    model.add_reactions([reaction])
    reaction.build_reaction_from_string("[c]: a --> b")
    assert len(model.metabolites) == 2
    assert model.metabolites.get_by_id("a[c]").compartment == "c"
    assert model.metabolites.get_by_id("b[c]").compartment == "c"
    assert model.reactions.R1.compartments == set(["c"])


def test_bounds_setter(model: Model) -> None:
    """Test reaction bounds setter."""
    rxn = model.reactions.get_by_id("PGI")
    with pytest.raises(ValueError):
        rxn.bounds = (1, 0)


def test_copy(model: Model) -> None:
    """Test reaction copying."""
    PGI = model.reactions.PGI
    copied = PGI.copy()
    assert PGI is not copied
    assert PGI._model is model
    assert copied._model is not model
    # The copy should refer to different metabolites and genes
    for met in copied.metabolites:
        assert met is not model.metabolites.get_by_id(met.id)
        assert met.model is not model
    for gene in copied.genes:
        assert gene is not model.genes.get_by_id(gene.id)
        assert gene.model is not model
    assert len(model.get_associated_groups(copied.id)) == 0


def test_iadd(model: Model) -> None:
    """Test in-place addition of reaction."""
    PGI = model.reactions.PGI
    EX_h2o = model.reactions.EX_h2o_e
    original_PGI_gpr = PGI.gene_reaction_rule
    PGI += EX_h2o
    assert PGI.gene_reaction_rule == original_PGI_gpr
    assert PGI.metabolites[model.metabolites.h2o_e] == -1.0
    # Original should not change
    assert EX_h2o.gene_reaction_rule == ""
    assert EX_h2o.metabolites[model.metabolites.h2o_e] == -1.0
    # Add a reaction not in the model
    new_reaction = Reaction("test")
    new_reaction.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    PGI += new_reaction
    assert PGI.gene_reaction_rule == original_PGI_gpr
    assert len(PGI.gene_reaction_rule) == 5
    # And vice versa
    new_reaction += PGI
    assert len(new_reaction.metabolites) == 5  # not
    assert len(new_reaction.genes) == 1
    assert new_reaction.gene_reaction_rule == original_PGI_gpr
    # Combine two GPRs
    model.reactions.ACKr += model.reactions.ACONTa
    expected_rule = "(b2296 or b3115 or b1849) and (b0118 or b1276)"
    assert model.reactions.ACKr.gene_reaction_rule == expected_rule
    assert len(model.reactions.ACKr.genes) == 5


def test_iadd_with_context(model: Model) -> None:
    """Test in-place addition of reaction is reversed with context."""
    PGI = model.reactions.PGI
    EX_h2o = model.reactions.EX_h2o_e
    original_PGI_gene_reaction_rule = PGI.gene_reaction_rule
    with model:
        PGI += EX_h2o
        assert PGI.gene_reaction_rule == original_PGI_gene_reaction_rule
        assert PGI.metabolites[model.metabolites.h2o_e] == -1.0
    assert PGI.gene_reaction_rule == original_PGI_gene_reaction_rule
    assert model.metabolites.h2o_e not in PGI.metabolites.keys()
    # Add a reaction not in the model
    new_reaction = Reaction("test")
    new_reaction.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    with model:
        PGI += new_reaction
    assert PGI.gene_reaction_rule == original_PGI_gene_reaction_rule
    assert len(PGI.gene_reaction_rule) == 5
    # Combine two GPRs
    expected_rule = "(b2296 or b3115 or b1849) and (b0118 or b1276)"
    old_rule = model.reactions.ACKr.gene_reaction_rule
    with model:
        model.reactions.ACKr += model.reactions.ACONTa
        assert model.reactions.ACKr.gene_reaction_rule == expected_rule
        assert len(model.reactions.ACKr.genes) == 5
    assert model.reactions.ACKr.gene_reaction_rule == old_rule
    assert old_rule != expected_rule
    assert len(model.reactions.ACKr.genes) == 3


def test_add(model: Model) -> None:
    """Test reaction addition to model."""
    # Not in place addition should work on a copy
    new = model.reactions.PGI + model.reactions.EX_h2o_e
    assert new._model is not model
    assert len(new.metabolites) == 3
    # The copy should refer to different metabolites and genes
    # This currently fails because add_metabolites does not copy.
    # Should that be changed?
    # for met in new.metabolites:
    #    assert met is not model.metabolites.get_by_id(met.id)
    #    assert met.model is not model
    for gene in new.genes:
        assert gene is not model.genes.get_by_id(gene.id)
        assert gene.model is not model


def test_radd(model: Model) -> None:
    """Test __radd__ for a reaction."""
    new = sum([model.reactions.PGI, model.reactions.EX_h2o_e])
    assert new._model is not model
    assert len(new.metabolites) == 3


def test_mul(model: Model) -> None:
    """Test scalar multiplication of factors with a reaction."""
    new = model.reactions.PGI * 2
    assert set(new.metabolites.values()) == {-2, 2}


def test_sub(model: Model) -> None:
    """Test reaction subtraction."""
    new = model.reactions.PGI - model.reactions.EX_h2o_e
    assert new._model is not model
    assert len(new.metabolites) == 3


def test_removal_from_model_retains_bounds(model: Model) -> None:
    """Test reaction removal from a model, retains its bounds."""
    model_cp = model.copy()
    reaction = model_cp.reactions.ACALD
    assert reaction.model == model_cp
    assert reaction.lower_bound == -1000.0
    assert reaction.upper_bound == 1000.0
    assert reaction._lower_bound == -1000.0
    assert reaction._upper_bound == 1000.0
    model_cp.remove_reactions([reaction])
    assert reaction.model is None
    assert reaction.lower_bound == -1000.0
    assert reaction.upper_bound == 1000.0
    assert reaction._lower_bound == -1000.0
    assert reaction._upper_bound == 1000.0


def test_set_bounds_scenario_1(model: Model) -> None:
    """Test reaction bounds setting for a scenario."""
    acald_reaction = model.reactions.ACALD
    assert acald_reaction.lower_bound == -1000.0
    assert acald_reaction.upper_bound == 1000.0
    assert acald_reaction.forward_variable.lb == 0.0
    assert acald_reaction.forward_variable.ub == 1000.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1000.0
    acald_reaction.bounds = (
        acald_reaction.lower_bound - 100,
        acald_reaction.lower_bound - 100,
    )
    assert acald_reaction.lower_bound == -1100.0
    assert acald_reaction.upper_bound == -1100.0
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 0
    assert acald_reaction.reverse_variable.lb == 1100.0
    assert acald_reaction.reverse_variable.ub == 1100.0
    acald_reaction.upper_bound = 100
    assert acald_reaction.lower_bound == -1100.0
    assert acald_reaction.upper_bound == 100
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 100
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1100.0


def test_set_bounds_scenario_2(model: Model) -> None:
    """Test reaction bounds setting for a scenario."""
    acald_reaction = model.reactions.ACALD
    assert acald_reaction.lower_bound == -1000.0
    assert acald_reaction.upper_bound == 1000.0
    assert acald_reaction.forward_variable.lb == 0.0
    assert acald_reaction.forward_variable.ub == 1000.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1000.0
    acald_reaction.bounds = (
        acald_reaction.upper_bound + 100,
        acald_reaction.upper_bound + 100,
    )
    assert acald_reaction.lower_bound == 1100.0
    assert acald_reaction.upper_bound == 1100.0
    assert acald_reaction.forward_variable.lb == 1100.0
    assert acald_reaction.forward_variable.ub == 1100.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 0
    acald_reaction.lower_bound = -100
    assert acald_reaction.lower_bound == -100.0
    assert acald_reaction.upper_bound == 1100.0
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 1100.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 100


def test_set_bounds_scenario_3(model: Model) -> None:
    """Test reaction bounds setting for a scenario."""
    reac = model.reactions.ACALD
    reac.bounds = (-10, -10)
    assert reac.lower_bound == -10
    assert reac.upper_bound == -10
    reac.bounds = (-9, -9)
    assert reac.lower_bound == -9
    assert reac.upper_bound == -9
    reac.bounds = (2, 2)
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2
    reac.bounds = (-10, -10)
    assert reac.lower_bound == -10
    assert reac.upper_bound == -10
    reac.bounds = (-11, -11)
    assert reac.lower_bound == -11
    assert reac.upper_bound == -11
    reac.upper_bound = 2
    assert reac.lower_bound == -11
    assert reac.upper_bound == 2


def test_set_bounds_scenario_4(model: Model) -> None:
    """Test reaction bounds setting for a scenario."""
    reac = model.reactions.ACALD
    reac.bounds = (2, 2)
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2
    assert reac.forward_variable.lb == 2
    assert reac.forward_variable.ub == 2
    reac.knock_out()
    reac.bounds = (-2, -2)
    assert reac.lower_bound == -2
    assert reac.upper_bound == -2
    assert reac.reverse_variable.lb == 2
    assert reac.reverse_variable.ub == 2


def test_set_upper_before_lower_bound_to_0(model: Model) -> None:
    """Test reaction bounds setting to zero."""
    model.reactions.GAPD.bounds = (0, 0)
    assert model.reactions.GAPD.lower_bound == 0
    assert model.reactions.GAPD.upper_bound == 0
    assert model.reactions.GAPD.forward_variable.lb == 0
    assert model.reactions.GAPD.forward_variable.ub == 0
    assert model.reactions.GAPD.reverse_variable.lb == 0
    assert model.reactions.GAPD.reverse_variable.ub == 0


def test_change_bounds(model: Model) -> None:
    """Test reaction bounds change."""
    reac = model.reactions.ACALD
    reac.bounds = (2, 2)
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2
    with model:
        reac.bounds = (5, 5)
        assert reac.lower_bound == 5
        assert reac.upper_bound == 5
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2


def test_make_irreversible(model: Model) -> None:
    """Test reaction irreversibility."""
    acald_reaction = model.reactions.ACALD
    assert acald_reaction.lower_bound == -1000.0
    assert acald_reaction.upper_bound == 1000.0
    assert acald_reaction.forward_variable.lb == 0.0
    assert acald_reaction.forward_variable.ub == 1000.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1000.0
    acald_reaction.lower_bound = 0
    assert acald_reaction.lower_bound == 0
    assert acald_reaction.upper_bound == 1000.0
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 1000.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 0
    acald_reaction.lower_bound = -100
    assert acald_reaction.lower_bound == -100.0
    assert acald_reaction.upper_bound == 1000.0
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 1000.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 100


def test_make_reversible(model: Model) -> None:
    """Test reaction reversibility."""
    pfk_reaction = model.reactions.PFK
    assert pfk_reaction.lower_bound == 0.0
    assert pfk_reaction.upper_bound == 1000.0
    assert pfk_reaction.forward_variable.lb == 0.0
    assert pfk_reaction.forward_variable.ub == 1000.0
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 0
    pfk_reaction.lower_bound = -100.0
    assert pfk_reaction.lower_bound == -100.0
    assert pfk_reaction.upper_bound == 1000.0
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 1000.0
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 100.0
    pfk_reaction.lower_bound = 0
    assert pfk_reaction.lower_bound == 0
    assert pfk_reaction.upper_bound == 1000.0
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 1000.0
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 0


def test_make_irreversible_irreversible_to_the_other_side(model: Model) -> None:
    """Test reaction irreversibility to irreversibility."""
    pfk_reaction = model.reactions.PFK
    assert pfk_reaction.lower_bound == 0.0
    assert pfk_reaction.upper_bound == 1000.0
    assert pfk_reaction.forward_variable.lb == 0.0
    assert pfk_reaction.forward_variable.ub == 1000.0
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 0
    pfk_reaction.bounds = (-100.0, -100.0)
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 0
    assert pfk_reaction.reverse_variable.lb == 100
    assert pfk_reaction.reverse_variable.ub == 100
    pfk_reaction.lower_bound = -1000.0
    assert pfk_reaction.lower_bound == -1000.0
    assert pfk_reaction.upper_bound == -100.0
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 0
    assert pfk_reaction.reverse_variable.lb == 100
    assert pfk_reaction.reverse_variable.ub == 1000.0


def test_make_lhs_irreversible_reversible(model: Model) -> None:
    """Test reaction LHS irreversibility to reversibility."""
    rxn = Reaction("test")
    rxn.add_metabolites({model.metabolites[0]: -1.0, model.metabolites[1]: 1.0})
    rxn.bounds = (-1000.0, -100)
    model.add_reactions([rxn])
    assert rxn.lower_bound == -1000.0
    assert rxn.upper_bound == -100.0
    assert rxn.forward_variable.lb == 0.0
    assert rxn.forward_variable.ub == 0.0
    assert rxn.reverse_variable.lb == 100.0
    assert rxn.reverse_variable.ub == 1000.0
    rxn.upper_bound = 666.0
    assert rxn.lower_bound == -1000.0
    assert rxn.upper_bound == 666.0
    assert rxn.forward_variable.lb == 0.0
    assert rxn.forward_variable.ub == 666
    assert rxn.reverse_variable.lb == 0.0
    assert rxn.reverse_variable.ub == 1000.0


def test_model_less_reaction(model: Model) -> None:
    """Test model without reactions."""
    model.slim_optimize()
    for reaction in model.reactions:
        assert isinstance(reaction.flux, float)
        assert isinstance(reaction.reduced_cost, float)
    for reaction in model.reactions:
        model.remove_reactions([reaction])
        with pytest.raises(RuntimeError):
            reaction.flux
        with pytest.raises(RuntimeError):
            reaction.reduced_cost


def test_knockout(model: Model) -> None:
    """Test reaction knockouts."""
    original_bounds = dict()
    for reaction in model.reactions:
        original_bounds[reaction.id] = (reaction.lower_bound, reaction.upper_bound)
        reaction.knock_out()
        assert reaction.lower_bound == 0
        assert reaction.upper_bound == 0
    for k, (lb, ub) in original_bounds.items():
        model.reactions.get_by_id(k).bounds = (lb, ub)
    for reaction in model.reactions:
        assert reaction.lower_bound == original_bounds[reaction.id][0]
        assert reaction.upper_bound == original_bounds[reaction.id][1]
    with model:
        for reaction in model.reactions:
            original_bounds[reaction.id] = (reaction.lower_bound, reaction.upper_bound)
            reaction.knock_out()
            assert reaction.lower_bound == 0
            assert reaction.upper_bound == 0
    for reaction in model.reactions:
        assert reaction.lower_bound == original_bounds[reaction.id][0]
        assert reaction.upper_bound == original_bounds[reaction.id][1]


def test_reaction_without_model() -> None:
    """Test reaction without model association."""
    r = Reaction("blub")
    assert r.flux_expression is None
    assert r.forward_variable is None
    assert r.reverse_variable is None


def test_weird_left_to_right_reaction_issue(tiny_toy_model: Model) -> None:
    """Test absurd left to right reaction."""
    d1 = tiny_toy_model.reactions.get_by_id("ex1")
    assert not d1.reversibility
    assert d1.lower_bound == -1000
    assert d1._lower_bound == -1000
    assert d1.upper_bound == 0
    assert d1._upper_bound == 0
    with tiny_toy_model:
        d1.knock_out()
        assert d1.lower_bound == 0
        assert d1._lower_bound == 0
        assert d1.upper_bound == 0
        assert d1._upper_bound == 0
    assert d1.lower_bound == -1000
    assert d1._lower_bound == -1000
    assert d1.upper_bound == 0
    assert d1._upper_bound == 0


def test_one_left_to_right_reaction_set_positive_ub(tiny_toy_model: Model) -> None:
    """Test left to right reaction with positive upper bound."""
    d1 = tiny_toy_model.reactions.get_by_id("ex1")
    assert d1.reverse_variable.lb == 0
    assert d1.reverse_variable.ub == 1000
    assert d1._lower_bound == -1000
    assert d1.lower_bound == -1000
    assert d1._upper_bound == 0
    assert d1.upper_bound == 0
    assert d1.forward_variable.lb == 0
    assert d1.forward_variable.ub == 0
    d1.upper_bound = 0.1
    assert d1.forward_variable.lb == 0
    assert d1.forward_variable.ub == 0.1
    assert d1.reverse_variable.lb == 0
    assert d1.reverse_variable.ub == 1000
    assert d1._lower_bound == -1000
    assert d1.upper_bound == 0.1
    assert d1._lower_bound == -1000
    assert d1.upper_bound == 0.1


def test_irrev_reaction_set_negative_lb(model: Model) -> None:
    """Test reaction irreversibility with negative lower bound."""
    assert not model.reactions.PFK.reversibility
    assert model.reactions.PFK.lower_bound == 0
    assert model.reactions.PFK.upper_bound == 1000.0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 1000.0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 0
    model.reactions.PFK.lower_bound = -1000
    assert model.reactions.PFK.lower_bound == -1000
    assert model.reactions.PFK.upper_bound == 1000.0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 1000.0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 1000


def test_twist_irrev_right_to_left_reaction_to_left_to_right(model: Model) -> None:
    """Test irreversibility reversal from right to left to left to right."""
    assert not model.reactions.PFK.reversibility
    assert model.reactions.PFK.lower_bound == 0
    assert model.reactions.PFK.upper_bound == 1000.0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 1000.0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 0
    model.reactions.PFK.bounds = (-1000, 0)
    assert model.reactions.PFK.lower_bound == -1000
    assert model.reactions.PFK.upper_bound == 0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 1000


def test_set_lb_higher_than_ub_sets_ub_to_new_lb(model: Model) -> None:
    """Test lower bound > upper bound makes upper bound to new lower bound."""
    for reaction in model.reactions:
        assert reaction.lower_bound <= reaction.upper_bound
        reaction.bounds = (reaction.upper_bound + 100, reaction.upper_bound + 100)
        assert reaction.lower_bound == reaction.upper_bound


def test_set_ub_lower_than_lb_sets_lb_to_new_ub(model: Model) -> None:
    """Test upper bound < lower bound makes lower bound to new upper bound."""
    for reaction in model.reactions:
        assert reaction.lower_bound <= reaction.upper_bound
        reaction.bounds = (reaction.lower_bound - 100, reaction.lower_bound - 100)
        assert reaction.lower_bound == reaction.upper_bound


def test_add_metabolites_combine_true(model: Model) -> None:
    """Test metabolite addition to reaction (with combine = True)."""
    test_metabolite = Metabolite("test")
    for reaction in model.reactions:
        reaction.add_metabolites({test_metabolite: -66}, combine=True)
        assert reaction.metabolites[test_metabolite] == -66
        assert (
            model.constraints["test"].get_linear_coefficients(
                [reaction.forward_variable]
            )[reaction.forward_variable]
            == -66
        )
        assert (
            model.constraints["test"].get_linear_coefficients(
                [reaction.reverse_variable]
            )[reaction.reverse_variable]
            == 66
        )
        already_included_metabolite = list(reaction.metabolites.keys())[0]
        previous_coefficient = reaction.get_coefficient(already_included_metabolite.id)
        reaction.add_metabolites({already_included_metabolite: 10}, combine=True)
        new_coefficient = previous_coefficient + 10
        assert reaction.metabolites[already_included_metabolite] == new_coefficient
        assert (
            model.constraints[already_included_metabolite.id].get_linear_coefficients(
                [reaction.forward_variable]
            )[reaction.forward_variable]
            == new_coefficient
        )
        assert (
            model.constraints[already_included_metabolite.id].get_linear_coefficients(
                [reaction.reverse_variable]
            )[reaction.reverse_variable]
            == -new_coefficient
        )


@pytest.mark.xfail(reason="non-deterministic test")
def test_add_metabolites_combine_false(model: Model) -> None:
    """Test metabolite addition to reaction (with combine = False)."""
    test_metabolite = Metabolite("test")
    for reaction in model.reactions:
        reaction.add_metabolites({test_metabolite: -66}, combine=False)
        assert reaction.metabolites[test_metabolite] == -66
        assert model.constraints["test"].expression.has(
            -66.0 * reaction.forward_variable
        )
        assert model.constraints["test"].expression.has(
            66.0 * reaction.reverse_variable
        )
        already_included_metabolite = list(reaction.metabolites.keys())[0]
        reaction.add_metabolites({already_included_metabolite: 10}, combine=False)
        assert reaction.metabolites[already_included_metabolite] == 10
        assert model.constraints[already_included_metabolite.id].expression.has(
            10 * reaction.forward_variable
        )
        assert model.constraints[already_included_metabolite.id].expression.has(
            -10 * reaction.reverse_variable
        )


def test_reaction_imul(model: Model) -> None:
    """Test in-place scalar factor multiplication to reaction."""
    with model:
        model.reactions.EX_glc__D_e *= 100
        assert (
            model.constraints.glc__D_e.expression.coeff(model.variables.EX_glc__D_e)
            == -100.0
        )
        assert model.reactions.EX_glc__D_e.reaction == "100.0 glc__D_e <=> "

    assert (
        model.constraints.glc__D_e.expression.coeff(model.variables.EX_glc__D_e) == -1.0
    )
    assert model.reactions.EX_glc__D_e.reaction == "glc__D_e <=> "

    with model:
        model.reactions.EX_glc__D_e *= -2
        assert model.reactions.EX_glc__D_e.bounds == (-1000.0, 10.0)
        assert model.reactions.EX_glc__D_e.reaction == " <=> 2.0 glc__D_e"

    assert model.reactions.EX_glc__D_e.bounds == (-10, 1000.0)
    assert model.reactions.EX_glc__D_e.reaction == "glc__D_e <=> "


# def test_pop(model):
#     pgi = model.reactions.PGI
#     g6p = model.metabolites.get_by_id("g6p_c")
#     f6p = model.metabolites.get_by_id("f6p_c")
#     g6p_expr = model.solver.constraints["g6p_c"].expression
#     g6p_coef = pgi.pop("g6p_c")
#     assert g6p not in pgi.metabolites
#     actual = model.solver.constraints[
#         "g6p_c"].expression.as_coefficients_dict()
#     expected = (g6p_expr - g6p_coef * pgi.flux_expression
#                 ).as_coefficients_dict()
#     assert actual == expected
#     assert pgi.metabolites[f6p] == 1
#
#     f6p_expr = model.solver.constraints["f6p_c"].expression
#     f6p_coef = pgi.pop(f6p)
#     assert f6p not in pgi.metabolites
#     assert model.solver.constraints[
#                "f6p_c"].expression.as_coefficients_dict() == (
#                f6p_expr - f6p_coef * pgi.flux_expression
#            ).as_coefficients_dict()


def test_remove_from_model(model: Model) -> None:
    """Test reaction removal from model."""
    pgi = model.reactions.PGI
    g6p = model.metabolites.g6p_c
    pgi_flux = model.optimize().fluxes["PGI"]
    assert abs(pgi_flux) > 1e-6

    with model:
        pgi.remove_from_model()
        assert pgi.model is None
        assert "PGI" not in model.reactions
        assert pgi.id not in model.variables
        assert pgi.reverse_id not in model.variables
        assert pgi not in g6p.reactions
        model.optimize()

    assert "PGI" in model.reactions
    assert pgi.id in model.variables
    assert pgi.reverse_id in model.variables
    assert pgi.forward_variable.problem is model.solver
    assert pgi in g6p.reactions
    assert g6p in pgi.metabolites
    assert np.isclose(pgi_flux, model.optimize().fluxes["PGI"])


def test_change_id_is_reflected_in_solver(model: Model) -> None:
    """Test reaction ID change reflection in solver."""
    for i, reaction in enumerate(model.reactions):
        old_reaction_id = reaction.id
        assert model.variables[old_reaction_id].name == old_reaction_id
        assert old_reaction_id in model.variables
        new_reaction_id = reaction.id + "_" + str(i)
        reaction.id = new_reaction_id
        assert reaction.id == new_reaction_id
        assert not (old_reaction_id in model.variables)
        assert reaction.id in model.variables
        assert reaction.reverse_id in model.variables
        name = model.variables[reaction.id].name
        assert name == reaction.id


def test_repr_html_(model: Model) -> None:
    """Test __repr_html__ functionality."""
    assert "<table>" in model.reactions[0]._repr_html_()


def test_compartment_changes(model: Model) -> None:
    """Test reaction compartment change."""
    rxn = model.reactions.EX_ac_e
    assert rxn.reactants[0].compartment in rxn.compartments
    rxn.reactants[0].compartment = "blub"
    assert rxn.reactants[0].compartment in rxn.compartments


def test_gpr_serialization(model: Model) -> None:
    """Verify that reactions GPRs are serialized compactly as str."""
    state = model.reactions[0].__getstate__()
    assert isinstance(state["_gpr"], str)


def test_gpr_serialization_backwards_compatibility(model: Model) -> None:
    """Verify that GPR serialization is backwards compatible."""
    state = model.reactions[0].__getstate__()
    print(state)
    state["gene_reaction_rule"] = state["_gpr"]  # old format
    del state["_gpr"]
    rxn = Reaction("test")
    rxn.__setstate__(state)
    assert isinstance(rxn.gpr, GPR)



"""Provides a function to find cyclic reactions in a metabolic model."""

from typing import TYPE_CHECKING, List, Optional, Tuple

import numpy as np
import optlang
from optlang.symbolics import Zero

from ..util import create_stoichiometric_matrix
from ..util import solver as sutil
from .helpers import normalize_cutoff


if TYPE_CHECKING:
    from cobra import Model


def _create_find_cyclic_reactions_problem(
    solver: "optlang.interface",
    s_int: np.ndarray,
    directions_int: np.ndarray,
    zero_cutoff: float,
    bound: float,
) -> Tuple["optlang.interface.Model", List["optlang.interface.Variable"]]:
    """Create an optimization problem to find cyclic reactions.

    This optimization problem includes only internal reactions, assumes
    steady-state constraints, and enforces reaction directions.
    """
    model = solver.Model()

    q_list = []
    for i in range(s_int.shape[1]):
        q = solver.Variable(
            name=f"q_{i}",
            lb=bound * min(0, directions_int[i, 0]),
            ub=bound * max(0, directions_int[i, 1]),
            problem=model,
        )
        q_list.append(q)
    model.add(q_list)

    for idx, row in enumerate(s_int):
        nnz_list = np.flatnonzero(np.abs(row) > zero_cutoff)
        if len(nnz_list) == 0:
            continue
        constraint = solver.Constraint(Zero, lb=0, ub=0, name=f"row_{idx}")
        model.add([constraint], sloppy=True)
        model.constraints[f"row_{idx}"].set_linear_coefficients(
            {q_list[i]: row[i] for i in nnz_list}
        )

    model.update()

    model.objective = solver.Objective(Zero)

    return model, q_list


def find_cyclic_reactions(
    model: "Model",
    zero_cutoff: Optional[float] = None,
    bound: float = 1e4,
    method: str = "optimized",
    required_stop_checks_num: int = 3,
) -> Tuple[List[str], List[Tuple[bool, bool]]]:
    """Find reactions, that can be in a loop in a steady state flux distribution.

    Parameters
    ----------
    model : cobra.Model
        The metabolic model to analyze.
    zero_cutoff : float, optional
        The cutoff value to consider a flux as zero.
        The default uses the `model.tolerance` (default None).
    bound : float, optional
        The bound for the reaction fluxes in the optimization problem.
        (default is 1e4).
    method : str, optional
        The method to use for finding cyclic reactions.
        Options are "optimized" (default) or "basic".
        See notes for details.
    required_stop_checks_num : int, optional
        This parameter is used only for the "optimized" method.
        The number of random checks to pass to prove that all cyclic
        reactions were found. (default is 3).

    Returns
    -------
    A tuple containing two lists:
        - A list of reaction IDs that can be part of a loop.
        - A list of tuples indicating the possible directions of
          reactions from the first list in the loop.
          Each tuple contains two boolean values: (can_be_negative, can_be_positive).

    Notes
    -----
    This function considers only the stoichiometric matrix and reaction directions.
    Any other constraints present in the model are ignored.
    * If a reaction is identified as cyclic, there may still be no feasible loop
      when taking into account all bounds and additional constraints.
    * However, if a reaction is identified as non-cyclic, it cannot participate
      in any loop in a steady-state flux distribution.

    The "basic" method for each reaction and direction checks if it can be a part of
    a loop by optimizing linear programming problem.

    The "optimized" method uses a faster randomized approach to firstly find all
    reactions that can be part of a loop and then checks their directions. This method
    usually works at least 2 times faster than the "basic" method.
    The `required_stop_checks_num` parameter is used to descrease the probability
    of missing some cyclic reactions.
    """

    if method not in ["optimized", "basic"]:
        raise ValueError(
            "The `method` parameter must be either 'optimized' or 'basic'."
        )

    if required_stop_checks_num < 1:
        raise ValueError(
            "The `required_stop_checks_num` parameter must be greater than 0."
        )

    zero_cutoff = normalize_cutoff(model, zero_cutoff)

    internal = [i for i, r in enumerate(model.reactions) if not r.boundary]
    s_int = create_stoichiometric_matrix(model)[:, np.array(internal)]
    n = s_int.shape[1]

    bounds_int = np.array([model.reactions[i].bounds for i in internal])
    directions_int = np.sign(bounds_int)

    lp, q_list = _create_find_cyclic_reactions_problem(
        model.problem, s_int, directions_int, zero_cutoff, bound
    )

    candidate_reactions = list(range(n))
    can_positive = [False] * n
    can_negative = [False] * n

    if method == "optimized":
        is_cyclic = [False] * n

        def set_reaction_weights():
            signs = 2 * np.random.randint(low=0, high=2, size=n) - 1
            weights = np.random.uniform(0.5, 1.0, size=n) * signs
            lp.objective.set_linear_coefficients(
                {q_list[i]: weights[i] for i in range(n) if not is_cyclic[i]}
            )

        set_reaction_weights()
        dir_order = ["min", "max"]
        stop_checks_num = 0
        cyclic_reactions_num = 0

        while cyclic_reactions_num < n and stop_checks_num < required_stop_checks_num:
            found_cyclic = False
            reverse_dir_order = False

            for dir in dir_order:
                lp.objective.direction = dir
                lp.optimize()

                sutil.check_solver_status(lp.status)

                if abs(lp.objective.value) > zero_cutoff:
                    remove_coef = {}
                    for i in range(n):
                        if not is_cyclic[i] and abs(q_list[i].primal) > zero_cutoff:
                            is_cyclic[i] = True
                            cyclic_reactions_num += 1
                            remove_coef[q_list[i]] = 0
                            if q_list[i].primal > zero_cutoff:
                                can_positive[i] = True
                            else:
                                can_negative[i] = True

                    found_cyclic = True
                    lp.objective.set_linear_coefficients(remove_coef)
                    stop_checks_num = 0
                    break

                reverse_dir_order = True

            if not found_cyclic:
                set_reaction_weights()
                stop_checks_num += 1

            if reverse_dir_order:
                dir_order.reverse()

        candidate_reactions = [i for i in range(n) if is_cyclic[i]]

    lp.objective = model.problem.Objective(Zero, direction="max")
    for i in candidate_reactions:
        for dir in ["min", "max"]:
            if (dir == "min" and can_negative[i]) or (dir == "max" and can_positive[i]):
                continue

            if directions_int[i, int(dir == "max")] == 0:
                continue

            lp.objective.set_linear_coefficients(
                {q_list[i]: 1.0 if dir == "max" else -1.0}
            )

            lp.optimize()
            sutil.check_solver_status(lp.status)

            if lp.objective.value > zero_cutoff:
                if dir == "max":
                    can_positive[i] = True
                else:
                    can_negative[i] = True

        lp.objective.set_linear_coefficients({q_list[i]: 0.0})

    cyclic_reactions = []
    cyclic_reactions_directions = []
    for i in range(n):
        if can_positive[i] or can_negative[i]:
            cyclic_reactions.append(model.reactions[internal[i]].id)
            cyclic_reactions_directions.append((can_negative[i], can_positive[i]))

    return cyclic_reactions, cyclic_reactions_directions



"""Test functions of gene.py ."""

from cobra.core import Model


def test_repr_html_(model: Model) -> None:
    """Test HTML represenation is correct for a gene.

    Parameters
    ----------
    model : cobra.Model
        The textbook model.

    """
    assert "<table>" in model.genes[0]._repr_html_()



"""Define global fixtures."""

import importlib.resources
from pathlib import Path
from pickle import load as _load
from typing import List, Tuple

import pytest

import cobra
from cobra import Metabolite, Model, Reaction, Solution
from cobra.io import read_sbml_model
from cobra.util import solver as sutil


data_dir = Path(__file__).parent / "data"
cobra_data_dir = importlib.resources.files(cobra.data)


def create_test_model(model_name: str = "salmonella") -> Model:
    """Return a cobra model for testing.

    Parameters
    ----------
    model_name: str
        One of 'ecoli', 'textbook', or 'salmonella', or the
        path to a pickled cobra.Model .

    Returns
    -------
    cobra.Model
        The cobra model.

    Raises
    ------
    OSError
        If no file is found at the path specified by `model_name`.

    """
    if model_name == "ecoli":
        ecoli_sbml = str((cobra_data_dir / "iJO1366.xml.gz").resolve())
        return read_sbml_model(ecoli_sbml)
    elif model_name == "textbook":
        textbook_sbml = str((cobra_data_dir / "textbook.xml.gz").resolve())
        return read_sbml_model(textbook_sbml)
    elif model_name == "mini":
        mini_sbml = str((data_dir / "mini_fbc2.xml").resolve())
        return read_sbml_model(mini_sbml)
    elif model_name == "salmonella":
        salmonella_pickle = str((data_dir / "salmonella.pickle").resolve())
        model_name = salmonella_pickle
    with open(model_name, mode="rb") as infile:
        return _load(infile)


@pytest.fixture(scope="session")
def data_directory() -> Path:
    """Provide session-level fixture for test data directory."""
    return data_dir


@pytest.fixture(scope="session")
def cobra_data_directory() -> Path:
    """Provide session-level fixture for cobra data directory."""
    return Path(cobra_data_dir)


@pytest.fixture(scope="session")
def empty_once() -> Model:
    """Provide session-level fixture for empty model."""
    return Model()


@pytest.fixture(scope="function")
def empty_model(empty_once: Model) -> Model:
    """Provide function-level fixture for empty model."""
    return empty_once.copy()


@pytest.fixture(scope="session")
def small_model() -> Model:
    """Provide session-level fixture for textbook model."""
    return create_test_model("textbook")


@pytest.fixture(scope="function")
def model(small_model: Model) -> Model:
    """Provide function-level fixture for textbook model."""
    return small_model.copy()


@pytest.fixture(scope="session")
def large_once() -> Model:
    """Provide session-level fixture for ecoli model."""
    return create_test_model("ecoli")


@pytest.fixture(scope="function")
def large_model(large_once: Model) -> Model:
    """Provide function-level fixture for ecoli model."""
    return large_once.copy()


@pytest.fixture(scope="session")
def medium_model() -> Model:
    """Provide session-level fixture for salmonella model."""
    return create_test_model("salmonella")


@pytest.fixture(scope="function")
def salmonella(medium_model: Model) -> Model:
    """Provide function-level fixture for salmonella model."""
    return medium_model.copy()


@pytest.fixture(scope="function")
def solved_model(data_directory: Path) -> Tuple[Solution, Model]:
    """Provide function-level fixture for solved textbook model."""
    model = create_test_model("textbook")
    with (data_directory / "textbook_solution.pickle").open(mode="rb") as infile:
        solution = _load(infile)
    return solution, model


@pytest.fixture(scope="session")
def tiny_toy_model() -> Model:
    """Provide session-level fixture for tiny toy model."""
    tiny = Model("Toy Model")
    m1 = Metabolite("M1")
    d1 = Reaction("ex1")
    d1.add_metabolites({m1: -1})
    d1.upper_bound = 0
    d1.lower_bound = -1000
    tiny.add_reactions([d1])
    tiny.objective = "ex1"
    return tiny


stable_optlang = ["glpk", "cplex", "gurobi"]
all_solvers = ["optlang-" + s for s in stable_optlang if s in sutil.solvers]


@pytest.fixture(params=all_solvers, scope="session")
def opt_solver(request: pytest.FixtureRequest) -> str:
    """Provide session-level fixture for parametrized optlang solver names."""
    return request.param


@pytest.fixture(scope="function")
def metabolites(model: Model, request: pytest.FixtureRequest) -> List[Metabolite]:
    """Provide function-level fixture for metabolite set based on `request`."""
    if request.param == "exchange":
        return [
            met
            for met in model.metabolites
            if met.compartment == "e" and "EX_" + met.id not in model.reactions
        ]
    elif request.param == "demand":
        return [
            met
            for met in model.metabolites
            if met.compartment == "c" and "DM_" + met.id not in model.reactions
        ]
    elif request.param == "sink":
        return [
            met
            for met in model.metabolites
            if met.compartment == "c" and "SK_" + met.id not in model.reactions
        ]
    else:
        raise ValueError("Unknown metabolites {}".format(request.param))



"""Test functionalities of model component pruning functions."""

from itertools import chain

from cobra.core import GPR, Metabolite, Model, Reaction
from cobra.manipulation import (
    knock_out_model_genes,
    prune_unused_metabolites,
    prune_unused_reactions,
    remove_genes,
)


def test_prune_unused_metabolites_output_type(model: Model) -> None:
    """Test the output type of unused metabolites pruning."""
    [model.reactions.get_by_id(x).remove_from_model() for x in ["RPI", "RPE", "GND"]]
    model_pruned, unused = prune_unused_metabolites(model)
    assert isinstance(model_pruned, Model)
    # test that the output contains metabolite objects
    assert isinstance(unused[0], Metabolite)


def test_prune_unused_metabolites_sanity(model: Model) -> None:
    """Test the sanity of unused metabolites pruning."""
    metabolite1 = model.metabolites.ru5p__D_c
    metabolite2 = model.metabolites.akg_e
    metabolite3 = model.metabolites.akg_c
    reactions = set(
        chain(metabolite1.reactions, metabolite2.reactions, metabolite3.reactions)
    )
    model.remove_reactions(reactions)
    model_pruned, unused = prune_unused_metabolites(model)
    assert metabolite1 in model.metabolites
    assert metabolite2 in model.metabolites
    assert metabolite3 in model.metabolites
    # test that the unused metabolites are not used in the model
    assert metabolite1 not in model_pruned.metabolites
    assert metabolite2 not in model_pruned.metabolites
    assert metabolite3 not in model_pruned.metabolites


def test_prune_unused_reactions_output_type(model: Model) -> None:
    """Test the output type of unused reactions pruning."""
    reaction = Reaction("foo")
    model.add_reactions([reaction])
    model_pruned, unused = prune_unused_reactions(model)
    assert isinstance(model_pruned, Model)
    # test that the output contains reaction objects
    assert isinstance(unused[0], Reaction)


def test_prune_unused_rxns_functionality(model: Model) -> None:
    """Test the sanity of unused reactions pruning."""
    for x in ["foo1", "foo2", "foo3"]:
        model.add_reactions([Reaction(x)])

    model_pruned, unused = prune_unused_reactions(model)
    assert "foo1" in model.reactions
    assert "foo2" in model.reactions
    assert "foo3" in model.reactions
    # test that the unused reactions are not used in the model
    assert "foo1" not in model_pruned.reactions
    assert "foo2" not in model_pruned.reactions
    assert "foo3" not in model_pruned.reactions


def test_gene_knockout(salmonella: Model) -> None:
    """Test gene knockout with and without context."""
    gene_list = ["STM1067", "STM0227"]
    dependent_reactions = [
        "3HAD100",
        "3HAD120",
        "3HAD121",
        "3HAD140",
        "3HAD141",
        "3HAD160",
        "3HAD161",
        "3HAD180",
        "3HAD181",
        "3HAD40",
        "3HAD60",
        "3HAD80",
        "T2DECAI",
    ]
    orig_gene_len = len(salmonella.genes)
    orig_bounds = salmonella.reactions.list_attr("bounds")
    with salmonella:
        expected_reactions = [
            salmonella.reactions.get_by_id(r) for r in dependent_reactions
        ]
        knocked_out_reactions = knock_out_model_genes(salmonella, gene_list)
        assert set(expected_reactions) == set(knocked_out_reactions)
    assert len(salmonella.genes) == orig_gene_len
    assert salmonella.reactions.list_attr("bounds") == orig_bounds
    with salmonella:
        expected_reactions = [salmonella.reactions.get_by_id("PGI")]
        knocked_out_reactions = knock_out_model_genes(salmonella, ["STM4221"])
        assert expected_reactions == knocked_out_reactions
    with salmonella:
        expected_reactions = [salmonella.reactions.get_by_id("PGI")]
        knocked_out_reactions = knock_out_model_genes(
            salmonella, [salmonella.genes.get_by_id("STM4221")]
        )
        assert expected_reactions == knocked_out_reactions
    with salmonella:
        expected_reactions = [salmonella.reactions.get_by_id("PGI")]
        knocked_out_reactions = knock_out_model_genes(
            salmonella, [salmonella.genes.index("STM4221")]
        )
        assert expected_reactions == knocked_out_reactions
    with salmonella:
        expected_reactions = [salmonella.reactions.get_by_id("4PEPTabcpp")]
        knocked_out_reactions = knock_out_model_genes(salmonella, ["STM1746.S"])
        assert expected_reactions == knocked_out_reactions
    knocked_out_reactions = knock_out_model_genes(salmonella, gene_list)
    assert len(knocked_out_reactions) == 13
    expected_reactions = [
        salmonella.reactions.get_by_id(r) for r in dependent_reactions
    ]
    assert set(knocked_out_reactions) == set(expected_reactions)
    knocked_out_reactions.extend(knock_out_model_genes(salmonella, ["STM4221"]))
    expected_reactions.append(salmonella.reactions.get_by_id("PGI"))
    assert set(knocked_out_reactions) == set(expected_reactions)
    # test computation when gene name is a subset of another
    test_model = Model()
    test_reaction_1 = Reaction("test1")
    test_reaction_1.gene_reaction_rule = "eggs or (spam and eggspam)"
    test_model.add_reactions([test_reaction_1])
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["eggs"])
        assert knocked_out_reactions == list()
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["eggs", "spam"])
        expected_reactions = [test_model.reactions.get_by_id("test1")]
        assert set(knocked_out_reactions) == set(expected_reactions)
    # test computation with nested boolean expression
    test_reaction_1.gene_reaction_rule = "g1 and g2 and (g3 or g4 or (g5 and g6))"
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["g3"])
        assert knocked_out_reactions == list()
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["g1"])
        assert knocked_out_reactions == [test_reaction_1]
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["g5"])
        assert knocked_out_reactions == list()
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["g3", "g4", "g5"])
        assert knocked_out_reactions == [test_reaction_1]
    # test computation when gene names are python expressions
    test_reaction_1.gene_reaction_rule = "g1 and (for or in)"
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["for", "in"])
        assert knocked_out_reactions == [test_reaction_1]
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["for"])
        assert knocked_out_reactions == list()
    test_reaction_1.gene_reaction_rule = "g1 and g2 and g2.conjugate"
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["g2"])
        assert knocked_out_reactions == [test_reaction_1]
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["g2.conjugate"])
        assert knocked_out_reactions == [test_reaction_1]
    test_reaction_1.gene_reaction_rule = "g1 and (try:' or 'except:1)"
    with test_model:
        knocked_out_reactions = knock_out_model_genes(test_model, ["try:'"])
        assert knocked_out_reactions == []
    with test_model:
        knocked_out_reactions = knock_out_model_genes(
            test_model, ["try:'", "'except:1"]
        )
        assert knocked_out_reactions == [test_reaction_1]


def test_remove_genes() -> None:
    """Test gene removal."""
    m = Model("test")
    m.add_reactions([Reaction("r" + str(i + 1)) for i in range(8)])
    assert len(m.reactions) == 8
    rxns = m.reactions
    rxns.r1.gene_reaction_rule = "(a and b) or (c and a)"
    rxns.r2.gene_reaction_rule = "(a and b and d and e)"
    rxns.r3.gene_reaction_rule = "(a and b) or (b and c)"
    rxns.r4.gene_reaction_rule = "(f and b) or (b and c)"
    rxns.r5.gene_reaction_rule = "x"
    rxns.r6.gene_reaction_rule = "y"
    rxns.r7.gene_reaction_rule = "x or     z"
    rxns.r8.gene_reaction_rule = ""
    assert "a" in m.genes
    assert "x" in m.genes
    remove_genes(m, ["a"], remove_reactions=False)
    assert "a" not in m.genes
    assert "x" in m.genes
    assert rxns.r1.gene_reaction_rule == ""
    assert rxns.r2.gene_reaction_rule == ""
    assert rxns.r3.gene_reaction_rule == "b and c"
    assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
    assert rxns.r5.gene_reaction_rule == "x"
    assert rxns.r6.gene_reaction_rule == "y"
    assert rxns.r7.genes == {m.genes.x, m.genes.z}
    assert rxns.r8.gene_reaction_rule == ""
    remove_genes(m, ["x"], remove_reactions=True)
    assert len(m.reactions) == 7
    assert "r5" not in m.reactions
    assert "x" not in m.genes
    assert rxns.r1.gene_reaction_rule == ""
    assert rxns.r2.gene_reaction_rule == ""
    assert rxns.r3.gene_reaction_rule == "b and c"
    assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
    assert rxns.r6.gene_reaction_rule == "y"
    assert rxns.r7.gene_reaction_rule == "z"
    assert rxns.r7.genes == {m.genes.z}
    assert rxns.r8.gene_reaction_rule == ""


def test_remove_genes_with_context() -> None:
    """Test gene removal is reversed in context."""
    m = Model("test")
    m.add_reactions([Reaction("r" + str(i + 1)) for i in range(8)])
    assert len(m.reactions) == 8
    rxns = m.reactions
    rxns.r1.gene_reaction_rule = "(a and b) or (c and a)"
    rxns.r2.gene_reaction_rule = "(a and b and d and e)"
    rxns.r3.gene_reaction_rule = "(a and b) or (b and c)"
    rxns.r4.gene_reaction_rule = "(f and b) or (b and c)"
    rxns.r5.gene_reaction_rule = "x"
    rxns.r6.gene_reaction_rule = "y"
    rxns.r7.gene_reaction_rule = "x or     z"
    rxns.r8.gene_reaction_rule = ""
    with m:
        remove_genes(m, ["a"], remove_reactions=False)
        assert "a" not in m.genes
        assert "x" in m.genes
        assert rxns.r1.gene_reaction_rule == ""
        assert rxns.r2.gene_reaction_rule == ""
        assert rxns.r3.gene_reaction_rule == "b and c"
        assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
        assert rxns.r5.gene_reaction_rule == "x"
        assert rxns.r6.gene_reaction_rule == "y"
        assert rxns.r7.genes == {m.genes.x, m.genes.z}
        assert rxns.r8.gene_reaction_rule == ""
    assert "a" in m.genes
    assert "x" in m.genes
    assert rxns.r1.gene_reaction_rule == "(a and b) or (c and a)"
    assert rxns.r2.gpr == GPR.from_string("(a and b and d and e)")
    assert rxns.r3.gene_reaction_rule == "(a and b) or (b and c)"
    assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
    assert rxns.r5.gene_reaction_rule == "x"
    assert rxns.r6.gene_reaction_rule == "y"
    assert rxns.r7.genes == {m.genes.x, m.genes.z}
    assert rxns.r8.gene_reaction_rule == ""
    with m:
        remove_genes(m, ["a"], remove_reactions=False)
        assert "a" not in m.genes
        remove_genes(m, ["x"], remove_reactions=True)
        assert len(m.reactions) == 7
        assert "r5" not in m.reactions
        assert "x" not in m.genes
        assert rxns.r1.gene_reaction_rule == ""
        assert rxns.r2.gene_reaction_rule == ""
        assert rxns.r3.gene_reaction_rule == "b and c"
        assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
        assert rxns.r6.gene_reaction_rule == "y"
        assert rxns.r7.gene_reaction_rule == "z"
        assert rxns.r7.genes == {m.genes.z}
        assert rxns.r8.gene_reaction_rule == ""
    assert "a" in m.genes
    assert "x" in m.genes
    assert len(m.reactions) == 8
    assert rxns.r1.gene_reaction_rule == "(a and b) or (c and a)"
    assert rxns.r2.gpr == GPR.from_string("(a and b and d and e)")
    assert rxns.r3.gene_reaction_rule == "(a and b) or (b and c)"
    assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
    assert rxns.r5.gene_reaction_rule == "x"
    assert rxns.r6.gene_reaction_rule == "y"
    assert rxns.r7.genes == {m.genes.x, m.genes.z}
    assert rxns.r8.gene_reaction_rule == ""



"""Test finding nullspace basis of matrix using Fast-SNP algorithm."""

from typing import List

import numpy as np

from cobra import Model
from cobra.flux_analysis.fast_snp import nullspace_fast_snp
from cobra.util import create_stoichiometric_matrix
from cobra.util import solver as sutil


def _validate_nullspace_basis(
    S: np.ndarray, directions: np.ndarray, basis: np.ndarray
) -> None:
    """Validate that the basis is a correct nullspace basis with given directions."""

    assert basis.shape[0] == S.shape[1]
    if basis.shape[1] == 0:
        return

    max_error = np.max(np.abs(S @ basis))
    assert max_error < 1e-6, "S @ basis is not zero"

    for v in basis.T:
        for i in range(len(v)):
            assert (
                directions[i, 0] == -1 or v[i] >= -1e-6
            ), f"Basis vector violates direction for index {i}"
            assert (
                directions[i, 1] == 1 or v[i] <= 1e-6
            ), f"Basis vector violates upper bound for index {i}"


def test_small_without_directions(all_solvers: List[str]) -> None:
    """Test nullspace_fast_snp on small example without directions."""

    S = np.array(
        [
            [1, -1, 0, 4, 2],
            [0, -1, 0, -1, -4],
            [3, 2, 1, 0, 2],
        ]
    )
    directions = np.array([[-1, 1]] * 5)

    basis = nullspace_fast_snp(sutil.solvers[all_solvers], S, directions)

    assert basis.shape == (5, 2)
    _validate_nullspace_basis(S, directions, basis)


def test_small_with_directions(all_solvers: List[str]) -> None:
    """Test nullspace_fast_snp on small example with directions."""

    S = np.array(
        [
            [1, -1, 0, 4, 2],
            [0, -1, 0, -1, -4],
            [3, 2, 1, 0, 2],
        ]
    )
    directions = np.array(
        [
            [0, 1],
            [0, 0],
            [-1, 0],
            [-1, 0],
            [-1, 1],
        ]
    )

    basis = nullspace_fast_snp(sutil.solvers[all_solvers], S, directions)

    assert basis.shape == (5, 1)
    _validate_nullspace_basis(S, directions, basis)


def test_random(all_solvers: List[str]) -> None:
    """Test nullspace_fast_snp on random matrices."""

    np.random.seed(42)
    for _ in range(20):
        m = np.random.randint(10, 15)
        n = np.random.randint(m + 1, m + 10)
        S = np.random.normal(size=(m, n))
        directions = np.array([[-1, 1]] * n)

        basis = nullspace_fast_snp(sutil.solvers[all_solvers], S, directions)
        _validate_nullspace_basis(S, directions, basis)


def test_stoichiometry_matrix(model: Model, all_solvers: List[str]) -> None:
    """Test nullspace_fast_snp on model's stoichiometric matrix and directions."""

    internal_reactions = [i for i, r in enumerate(model.reactions) if not r.boundary]
    S_int = create_stoichiometric_matrix(model)[:, np.array(internal_reactions)]

    no_directions = np.array([[-1, 1]] * S_int.shape[1])
    basis_without_directions = nullspace_fast_snp(
        sutil.solvers[all_solvers], S_int, no_directions
    )
    _validate_nullspace_basis(S_int, no_directions, basis_without_directions)

    directions = np.sign(
        np.array([model.reactions[i].bounds for i in internal_reactions])
    )
    basis_with_directions = nullspace_fast_snp(
        sutil.solvers[all_solvers], S_int, directions
    )
    _validate_nullspace_basis(S_int, directions, basis_with_directions)

    assert basis_with_directions.shape[1] <= basis_without_directions.shape[1]



"""Define the Model class."""

import logging
from copy import copy, deepcopy
from functools import partial
from types import ModuleType
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple, Union
from warnings import warn

import optlang
from optlang.symbolics import Basic, Zero

from ..medium import find_boundary_types, find_external_compartment, sbo_terms
from ..util.context import HistoryManager, get_context, resettable
from ..util.solver import (
    add_cons_vars_to_problem,
    assert_optimal,
    check_solver,
    interface_to_str,
    remove_cons_vars_from_problem,
    set_objective,
)
from ..util.util import AutoVivification, format_long_string
from .configuration import Configuration
from .dictlist import DictList
from .gene import Gene
from .group import Group
from .metabolite import Metabolite
from .object import Object
from .reaction import Reaction
from .solution import get_solution


if TYPE_CHECKING:
    import pandas as pd
    from optlang.container import Container

    from cobra import Solution
    from cobra.summary import ModelSummary
    from cobra.util.solver import CONS_VARS

logger = logging.getLogger(__name__)
configuration = Configuration()


class Model(Object):
    """Class representation for a cobra model.

    Parameters
    ----------
    id_or_model: str or Model, optional
        String to use as model id, or actual model to base new model one.
        If string, it is used as id. If model, a new model object is
        instantiated with the same properties as the original model (default None).
    name: str, optional
        Human readable string to be model description (default None).

    Attributes
    ----------
    reactions : DictList
        A DictList where the key is the reaction identifier and the value a
        Reaction
    metabolites : DictList
        A DictList where the key is the metabolite identifier and the value a
        Metabolite
    genes : DictList
        A DictList where the key is the gene identifier and the value a
        Gene
    groups : DictList
        A DictList where the key is the group identifier and the value a
        Group
    """

    def __init__(
        self, id_or_model: Union[str, "Model", None] = None, name: Optional[str] = None
    ) -> None:
        """Initialize the Model."""
        if isinstance(id_or_model, Model):
            Object.__init__(self, name=name)
            self.__setstate__(id_or_model.__dict__)
            self._solver = id_or_model.solver
        else:
            Object.__init__(self, id_or_model, name=name)
            self.genes: DictList[Gene] = DictList()
            self.reactions: DictList[Reaction] = DictList()
            self.metabolites: DictList[Metabolite] = DictList()
            self.groups: DictList[Group] = DictList()
            # genes based on their ids {Gene.id: Gene}
            self._compartments = {}
            self._contexts = []

            # from cameo ...

            # if not hasattr(self, '_solver'):  # backwards compatibility
            # with older cobrapy pickles?

            interface = check_solver(configuration.solver)
            self._solver = interface.Model()
            self._solver.objective = interface.Objective(Zero)
            self._populate_solver(self.reactions, self.metabolites)

            self._tolerance = None
            self.tolerance = configuration.tolerance

    def __setstate__(self, state: Dict) -> None:
        """Make sure all cobra.Objects in the model point to the model.

        Parameters
        ----------
        state: dict
        """
        self.__dict__.update(state)
        for y in ["reactions", "genes", "metabolites"]:
            for x in getattr(self, y):
                x._model = self
        if not hasattr(self, "name"):
            self.name = None

    def __getstate__(self) -> Dict:
        """Get state for serialization.

        Ensures that the context stack is cleared prior to serialization,
        since partial functions cannot be pickled reliably.

        Returns
        -------
        odict: Dict
            A dictionary of state, based on self.__dict__.
        """
        odict = self.__dict__.copy()
        odict["_contexts"] = []
        return odict

    @property
    def solver(self) -> "optlang.interface.Model":
        """Get the attached solver instance.

        The associated the solver object, which manages the interaction with
        the associated solver, e.g. glpk.

        This property is useful for accessing the optimization problem
        directly and to define additional non-metabolic constraints.

        Examples
        --------
        >>> from cobra.io import load_model
        >>> model = load_model("textbook")
        >>> new = model.problem.Constraint(model.objective.expression, lb=0.99)
        >>> model.solver.add(new)
        """
        return self._solver

    @solver.setter
    @resettable
    def solver(self, value: Union[str, ModuleType]) -> None:
        """Set the attached solver instance.

        The associated the solver object, which manages the interaction with
        the associated solver, e.g. glpk.

        This property is useful for accessing the optimization problem
        directly and to define additional non-metabolic constraints.

        Parameters
        ----------
        value: ModuleType or str
            The solver to set, which will be checked by `check_solver()`.
        """
        interface = check_solver(value)

        # Do nothing if the solver did not change
        if self.problem == interface:
            return
        self._solver = interface.Model.clone(self._solver)

    @property
    def tolerance(self) -> float:
        """Get the tolerance.

        Returns
        -------
        float
            The tolerance of the mdoel.
        """
        return self._tolerance

    @tolerance.setter
    def tolerance(self, value: float) -> None:
        """Set the tolerance.

        Parameters
        ----------
        value: float
            Value to set tolerance.
        """
        solver_tolerances = self.solver.configuration.tolerances

        try:
            solver_tolerances.feasibility = value
        except AttributeError:
            logger.info(
                f"The current solver interface {interface_to_str(self.problem)} "
                f"doesn't support setting the feasibility tolerance."
            )

        try:
            solver_tolerances.optimality = value
        except AttributeError:
            logger.info(
                f"The current solver interface {interface_to_str(self.problem)} "
                f"doesn't support setting the optimality tolerance."
            )

        try:
            solver_tolerances.integrality = value
        except AttributeError:
            logger.info(
                f"The current solver interface {interface_to_str(self.problem)} "
                f"doesn't support setting the integrality tolerance."
            )

        self._tolerance = value

    @property
    def compartments(self) -> Dict:
        """Return all metabolites' compartments.

        Returns
        -------
        dict
            A dictionary of metabolite compartments, where the keys are the short
            version (one letter version) of the compartments, and the values are the
            full names (if they exist).
        """
        return {
            met.compartment: self._compartments.get(met.compartment, "")
            for met in self.metabolites
            if met.compartment is not None
        }

    @compartments.setter
    def compartments(self, value: Dict) -> None:
        """Get or set the dictionary of current compartment descriptions.

        Assigning a dictionary to this property updates the model's
        dictionary of compartment descriptions with the new values.

        Parameters
        ----------
        value : dict
            Dictionary mapping compartments abbreviations to full names.

        Examples
        --------
        >>> from cobra.io import load_model
        >>> model = load_model("textbook")
        >>> model.compartments = {'c': 'the cytosol'}
        >>> model.compartments
        {'c': 'the cytosol', 'e': 'extracellular'}
        """
        self._compartments.update(value)

    @property
    def medium(self) -> Dict[str, float]:
        """Get the constraints on the model exchanges.

        `model.medium` returns a dictionary of the bounds for each of the
        boundary reactions, in the form of `{rxn_id: bound}`, where `bound`
        specifies the absolute value of the bound in direction of metabolite
        creation (i.e., lower_bound for `met <--`, upper_bound for `met -->`)

        Returns
        -------
        Dict[str, float]
            A dictionary with rxn.id (str) as key, bound (float) as value.
        """

        def is_active(reaction: Reaction) -> bool:
            """Determine if boundary reaction permits flux towards creating metabolites.

            Parameters
            ----------
            reaction: cobra.Reaction

            Returns
            -------
            bool
                True if reaction produces metaoblites and has upper_bound above 0
                or if reaction consumes metabolites and has lower_bound below 0 (so
                could be reversed).
            """
            return (bool(reaction.products) and (reaction.upper_bound > 0)) or (
                bool(reaction.reactants) and (reaction.lower_bound < 0)
            )

        def get_active_bound(reaction: Reaction) -> float:
            """For an active boundary reaction, return the relevant bound.

            Parameters
            ----------
            reaction: cobra.Reaction

            Returns
            -------
            float:
                upper or minus lower bound, depenending if the reaction produces or
                consumes metaoblties.
            """
            if reaction.reactants:
                return -reaction.lower_bound
            elif reaction.products:
                return reaction.upper_bound

        return {
            rxn.id: get_active_bound(rxn) for rxn in self.exchanges if is_active(rxn)
        }

    @medium.setter
    def medium(self, medium: Dict[str, float]) -> None:
        """Set the constraints on the model exchanges.

        `model.medium` returns a dictionary of the bounds for each of the
        boundary reactions, in the form of `{rxn_id: rxn_bound}`, where `rxn_bound`
        specifies the absolute value of the bound in direction of metabolite
        creation (i.e., lower_bound for `met <--`, upper_bound for `met -->`)

        Parameters
        ----------
        medium: dict
            The medium to initialize. medium should be a dictionary defining
            `{rxn_id: bound}` pairs.
        """

        def set_active_bound(reaction: Reaction, bound: float) -> None:
            """Set active bound.

            Parameters
            ----------
            reaction: cobra.Reaction
                Reaction to set
            bound: float
                Value to set bound to. The bound is reversed and set as lower bound
                if reaction has reactants (metabolites that are consumed). If reaction
                has reactants, it seems the upper bound won't be set.
            """
            if reaction.reactants:
                reaction.lower_bound = -bound
            elif reaction.products:
                reaction.upper_bound = bound

        # Set the given media bounds
        media_rxns = []
        exchange_rxns = frozenset(self.exchanges)
        for rxn_id, rxn_bound in medium.items():
            rxn = self.reactions.get_by_id(rxn_id)
            if rxn not in exchange_rxns:
                logger.warning(
                    f"{rxn.id} does not seem to be an an exchange reaction. "
                    f"Applying bounds anyway."
                )
            media_rxns.append(rxn)
            # noinspection PyTypeChecker
            set_active_bound(rxn, rxn_bound)

        frozen_media_rxns = frozenset(media_rxns)

        # Turn off reactions not present in media
        for rxn in exchange_rxns - frozen_media_rxns:
            is_export = rxn.reactants and not rxn.products
            set_active_bound(
                rxn, min(0.0, -rxn.lower_bound if is_export else rxn.upper_bound)
            )

    def copy(self) -> "Model":
        """Provide a partial 'deepcopy' of the Model.

        All the Metabolite, Gene, and Reaction objects are created anew but
        in a faster fashion than deepcopy.

        Returns
        -------
        cobra.Model: new model copy
        """
        new = self.__class__()
        do_not_copy_by_ref = {
            "metabolites",
            "reactions",
            "genes",
            "notes",
            "annotation",
            "groups",
        }
        for attr in self.__dict__:
            if attr not in do_not_copy_by_ref:
                new.__dict__[attr] = self.__dict__[attr]
        new.notes = deepcopy(self.notes)
        new.annotation = deepcopy(self.annotation)

        new.metabolites = DictList()
        do_not_copy_by_ref = {"_reaction", "_model"}
        for metabolite in self.metabolites:
            new_met = metabolite.__class__()
            for attr, value in metabolite.__dict__.items():
                if attr not in do_not_copy_by_ref:
                    new_met.__dict__[attr] = copy(value) if attr == "formula" else value
            new_met._model = new
            new.metabolites.append(new_met)

        new.genes = DictList()
        for gene in self.genes:
            new_gene = gene.__class__(None)
            for attr, value in gene.__dict__.items():
                if attr not in do_not_copy_by_ref:
                    new_gene.__dict__[attr] = (
                        copy(value) if attr == "formula" else value
                    )
            new_gene._model = new
            new.genes.append(new_gene)

        new.reactions = DictList()
        do_not_copy_by_ref = {"_model", "_metabolites", "_genes"}
        for reaction in self.reactions:
            new_reaction = reaction.__class__()
            for attr, value in reaction.__dict__.items():
                if attr not in do_not_copy_by_ref:
                    new_reaction.__dict__[attr] = copy(value)
            new_reaction._model = new
            new.reactions.append(new_reaction)
            # update awareness
            for metabolite, stoic in reaction._metabolites.items():
                new_met = new.metabolites.get_by_id(metabolite.id)
                new_reaction._metabolites[new_met] = stoic
                new_met._reaction.add(new_reaction)
            new_reaction.update_genes_from_gpr()

        new.groups = DictList()
        do_not_copy_by_ref = {"_model", "_members"}
        # Groups can be members of other groups. We initialize them first and
        # then update their members.
        for group in self.groups:
            new_group: Group = group.__class__(group.id)
            for attr, value in group.__dict__.items():
                if attr not in do_not_copy_by_ref:
                    new_group.__dict__[attr] = copy(value)
            new_group._model = new
            new.groups.append(new_group)
        for group in self.groups:
            new_group = new.groups.get_by_id(group.id)
            # update awareness, as in the reaction copies
            new_objects = []
            for member in group.members:
                if isinstance(member, Metabolite):
                    new_object = new.metabolites.get_by_id(member.id)
                elif isinstance(member, Reaction):
                    new_object = new.reactions.get_by_id(member.id)
                elif isinstance(member, Gene):
                    new_object = new.genes.get_by_id(member.id)
                elif isinstance(member, Group):
                    new_object = new.groups.get_by_id(member.id)
                else:
                    raise TypeError(
                        f"The group member {member!r} is unexpectedly not a "
                        f"metabolite, reaction, gene, nor another group."
                    )
                new_objects.append(new_object)
            new_group.add_members(new_objects)

        try:
            new._solver = deepcopy(self.solver)
            # Cplex has an issue with deep copies
        except Exception:  # pragma: no cover
            new._solver = copy(self.solver)  # pragma: no cover

        # it doesn't make sense to retain the context of a copied model so
        # assign a new empty context
        new._contexts = []

        return new

    def add_metabolites(self, metabolite_list: Union[List, Metabolite]) -> None:
        """Add new metabolites to a model.

        Will add a list of metabolites to the model object and add new
        constraints accordingly.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        metabolite_list : list or Metabolite.
            A list of `cobra.core.Metabolite` objects. If it isn't an iterable
            container, the metabolite will be placed into a list.

        """
        if not hasattr(metabolite_list, "__iter__"):
            metabolite_list = [metabolite_list]
        if len(metabolite_list) == 0:
            return None

        # First check whether the metabolites exist in the model
        metabolite_list = [x for x in metabolite_list if x.id not in self.metabolites]

        bad_ids = [
            m for m in metabolite_list if not isinstance(m.id, str) or len(m.id) < 1
        ]
        if len(bad_ids) != 0:
            raise ValueError(f"invalid identifiers in {repr(bad_ids)}")

        for x in metabolite_list:
            x._model = self
        self.metabolites += metabolite_list

        # from cameo ...
        to_add = []
        for met in metabolite_list:
            if met.id not in self.constraints:
                constraint = self.problem.Constraint(Zero, name=met.id, lb=0, ub=0)
                to_add += [constraint]

        self.add_cons_vars(to_add)

        context = get_context(self)
        if context:
            context(partial(self.metabolites.__isub__, metabolite_list))
            for x in metabolite_list:
                # Do we care?
                context(partial(setattr, x, "_model", None))

    def remove_metabolites(
        self, metabolite_list: Union[List, Metabolite], destructive: bool = False
    ) -> None:
        """Remove a list of metabolites from the the object.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        metabolite_list : list or Metaoblite
            A list of `cobra.core.Metabolite` objects. If it isn't an iterable
            container, the metabolite will be placed into a list.

        destructive : bool, optional
            If False then the metabolite is removed from all
            associated reactions.  If True then all associated
            reactions are removed from the Model (default False).
        """
        if not hasattr(metabolite_list, "__iter__"):
            metabolite_list = [metabolite_list]
        # Make sure metabolites exist in model
        metabolite_list = [x for x in metabolite_list if x.id in self.metabolites]
        for x in metabolite_list:
            x._model = None

            # remove reference to the metabolite in all groups
            associated_groups = self.get_associated_groups(x)
            for group in associated_groups:
                group.remove_members(x)

            if not destructive:
                for the_reaction in list(x._reaction):  # noqa W0212
                    the_coefficient = the_reaction._metabolites[x]  # noqa W0212
                    the_reaction.subtract_metabolites({x: the_coefficient})

            else:
                for x2 in list(x._reaction):  # noqa W0212
                    x2.remove_from_model()

        self.metabolites -= metabolite_list

        to_remove = [self.solver.constraints[m.id] for m in metabolite_list]
        self.remove_cons_vars(to_remove)

        context = get_context(self)
        if context:
            context(partial(self.metabolites.__iadd__, metabolite_list))
            for x in metabolite_list:
                context(partial(setattr, x, "_model", self))

    def add_boundary(
        self,
        metabolite: Metabolite,
        type: str = "exchange",
        reaction_id: Optional[str] = None,
        lb: Optional[float] = None,
        ub: Optional[float] = None,
        sbo_term: Optional[str] = None,
    ) -> Reaction:
        """
        Add a boundary reaction for a given metabolite.

        There are three different types of pre-defined boundary reactions:
        exchange, demand, and sink reactions.
        An exchange reaction is a reversible, unbalanced reaction that adds
        to or removes an extracellular metabolite from the extracellular
        compartment.
        A demand reaction is an irreversible reaction that consumes an
        intracellular metabolite.
        A sink is similar to an exchange but specifically for intracellular
        metabolites, i.e., a reversible reaction that adds or removes an
        intracellular metabolite.

        If you set the reaction `type` to something else, you must specify the
        desired identifier of the created reaction along with its upper and
        lower bound. The name will be given by the metabolite name and the
        given `type`.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        metabolite : cobra.Metabolite
            Any given metabolite. The compartment is not checked but you are
            encouraged to stick to the definition of exchanges and sinks.
        type : {"exchange", "demand", "sink"}
            Using one of the pre-defined reaction types is easiest. If you
            want to create your own kind of boundary reaction choose
            any other string, e.g., 'my-boundary' (default "exchange").
        reaction_id : str, optional
            The ID of the resulting reaction. This takes precedence over the
            auto-generated identifiers but beware that it might make boundary
            reactions harder to identify afterwards when using `model.boundary`
            or specifically `model.exchanges` etc. (default None).
        lb : float, optional
            The lower bound of the resulting reaction (default None).
        ub : float, optional
            The upper bound of the resulting reaction (default None).
        sbo_term : str, optional
            A correct SBO term is set for the available types. If a custom
            type is chosen, a suitable SBO term should also be set (default None).

        Returns
        -------
        cobra.Reaction
            The created boundary reaction.

        Examples
        --------
        >>> from cobra.io load_model
        >>> model = load_model("textbook")
        >>> demand = model.add_boundary(model.metabolites.atp_c, type="demand")
        >>> demand.id
        'DM_atp_c'
        >>> demand.name
        'ATP demand'
        >>> demand.bounds
        (0, 1000.0)
        >>> demand.build_reaction_string()
        'atp_c --> '

        """
        ub = configuration.upper_bound if ub is None else ub
        lb = configuration.lower_bound if lb is None else lb
        types = {
            "exchange": ("EX", lb, ub, sbo_terms["exchange"]),
            "demand": ("DM", 0, ub, sbo_terms["demand"]),
            "sink": ("SK", lb, ub, sbo_terms["sink"]),
        }
        if type == "exchange":
            external = find_external_compartment(self)
            if metabolite.compartment != external:
                raise ValueError(
                    f"The metabolite is not an external metabolite (compartment is "
                    f"`{metabolite.compartment}` but should be `{external}`). "
                    f"Did you mean to add a demand or sink? If not, either change"
                    f" its compartment or rename the model compartments to fix this."
                )
        if type in types:
            prefix, lb, ub, default_term = types[type]
            if reaction_id is None:
                reaction_id = f"{prefix}_{metabolite.id}"
            if sbo_term is None:
                sbo_term = default_term
        if reaction_id is None:
            raise ValueError(
                "Custom types of boundary reactions require a custom "
                "identifier. Please set the `reaction_id`."
            )
        if reaction_id in self.reactions:
            # It already exists so just retrieve it.
            logger.info(f"Boundary reaction '{reaction_id}' already exists.")
            return self.reactions.get_by_id(reaction_id)
        name = f"{metabolite.name} {type}"
        rxn = Reaction(id=reaction_id, name=name, lower_bound=lb, upper_bound=ub)
        rxn.add_metabolites({metabolite: -1})
        if sbo_term:
            rxn.annotation["sbo"] = sbo_term
        self.add_reactions([rxn])
        return rxn

    def add_reactions(self, reaction_list: Iterable[Reaction]) -> None:
        """Add reactions to the model.

        Reactions with identifiers identical to a reaction already in the
        model are ignored.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        reaction_list : list
            A list of `cobra.Reaction` objects
        """

        def existing_filter(rxn: Reaction) -> bool:
            """Check if the reaction does not exists in the model.

            Parameters
            ----------
            rxn: cobra.Reaction

            Returns
            -------
            bool
                False if reaction exists, True if it doesn't.
                If the reaction exists, will log a warning.
            """
            if rxn.id in self.reactions:
                logger.warning(f"Ignoring reaction '{rxn.id}' since it already exists.")
                return False
            return True

        # First check whether the reactions exist in the model.
        pruned = DictList(filter(existing_filter, reaction_list))

        context = get_context(self)

        # Add reactions. Also take care of genes and metabolites in the loop.
        for reaction in pruned:
            reaction._model = self
            if context:
                context(partial(setattr, reaction, "_model", None))
            # Build a `list()` because the dict will be modified in the loop.
            for metabolite in list(reaction.metabolites):
                # TODO: Maybe this can happen with
                #  Reaction.add_metabolites(combine=False)
                # TODO: Should we add a copy of the metabolite instead?
                if metabolite not in self.metabolites:
                    self.add_metabolites(metabolite)
                # A copy of the metabolite exists in the model, the reaction
                # needs to point to the metabolite in the model.
                else:
                    # FIXME: Modifying 'private' attributes is horrible.
                    stoichiometry = reaction._metabolites.pop(metabolite)
                    model_metabolite = self.metabolites.get_by_id(metabolite.id)
                    reaction._metabolites[model_metabolite] = stoichiometry
                    model_metabolite._reaction.add(reaction)
                    if context:
                        context(partial(model_metabolite._reaction.remove, reaction))
            reaction.update_genes_from_gpr()

        self.reactions += pruned

        if context:
            context(partial(self.reactions.__isub__, pruned))

        # from cameo ...
        self._populate_solver(pruned)

    def remove_reactions(
        self,
        reactions: Union[str, Reaction, List[Union[str, Reaction]]],
        remove_orphans: bool = False,
    ) -> None:
        """Remove reactions from the model.

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        reactions : list or reaction or str
            A list with reactions (`cobra.Reaction`), or their id's, to remove.
            Reaction will be placed in a list. Str will be placed in a list and used to
            find the reaction in the model.
        remove_orphans : bool, optional
            Remove orphaned genes and metabolites from the model as
            well (default False).
        """
        if isinstance(reactions, str) or hasattr(reactions, "id"):
            warn("need to pass in a list")
            reactions = [reactions]

        context = get_context(self)

        for reaction in reactions:
            # Make sure the reaction is in the model
            try:
                reaction = self.reactions[self.reactions.index(reaction)]
            except ValueError:
                warn(f"{reaction} not in {self}")

            else:
                forward = reaction.forward_variable
                reverse = reaction.reverse_variable

                if context:
                    obj_coef = reaction.objective_coefficient

                    if obj_coef != 0:
                        context(
                            partial(
                                self.solver.objective.set_linear_coefficients,
                                {forward: obj_coef, reverse: -obj_coef},
                            )
                        )

                    context(partial(self._populate_solver, [reaction]))
                    context(partial(setattr, reaction, "_model", self))
                    context(partial(self.reactions.add, reaction))

                self.remove_cons_vars([forward, reverse])
                self.reactions.remove(reaction)
                reaction._model = None

                for met in reaction._metabolites:
                    if reaction in met._reaction:
                        met._reaction.remove(reaction)
                        if context:
                            context(partial(met._reaction.add, reaction))
                        if remove_orphans and len(met._reaction) == 0:
                            self.remove_metabolites(met)

                for gene in reaction._genes:
                    if reaction in gene._reaction:
                        gene._reaction.remove(reaction)
                        if context:
                            context(partial(gene._reaction.add, reaction))

                        if remove_orphans and len(gene._reaction) == 0:
                            self.genes.remove(gene)
                            if context:
                                context(partial(self.genes.add, gene))

                # remove reference to the reaction in all groups
                associated_groups = self.get_associated_groups(reaction)
                for group in associated_groups:
                    group.remove_members(reaction)

    def add_groups(self, group_list: Union[str, Group, List[Group]]) -> None:
        """Add groups to the model.

        Groups with identifiers identical to a group already in the model are
        ignored.

        If any group contains members that are not in the model, these members
        are added to the model as well. Only metabolites, reactions, and genes
        can have groups.

        Parameters
        ----------
        group_list : list or str or Group
            A list of `cobra.Group` objects to add to the model. Can also be a single
            group or a string representing group id. If the input is not a list, a
            warning is raised.
        """

        def existing_filter(new_group: Group) -> bool:
            """Check if the group does not exist.

            Parameters
            ----------
            new_group: cobra.Group
                Group to check.

            Returns
            -------
            bool
                False if the group already exists, True if it doesn't.
            """
            if new_group.id in self.groups:
                logger.warning(
                    f"Ignoring group '{new_group.id}'" f" since it already exists."
                )
                return False
            return True

        if isinstance(group_list, str) or hasattr(group_list, "id"):
            warn("need to pass in a list")
            group_list = [group_list]

        pruned = DictList(filter(existing_filter, group_list))

        for group in pruned:
            group._model = self
            for member in group.members:
                # If the member is not associated with the model, add it
                if isinstance(member, Metabolite) and member not in self.metabolites:
                    self.add_metabolites([member])
                if isinstance(member, Reaction) and member not in self.reactions:
                    self.add_reactions([member])
                # TODO(midnighter): `add_genes` method does not exist.
                # if isinstance(member, Gene):
                #     if member not in self.genes:
                #         self.add_genes([member])

            self.groups += [group]

    def remove_groups(self, group_list: Union[str, Group, List[Group]]) -> None:
        """Remove groups from the model.

        Members of each group are not removed
        from the model (i.e. metabolites, reactions, and genes in the group
        stay in the model after any groups containing them are removed).

        Parameters
        ----------
        group_list : list or str or Group
            A list of `cobra.Group` objects to remove from the model. Can also be a
            single group or a string representing group id. If the input is not a list,
             a warning is raised.
        """
        if isinstance(group_list, str) or hasattr(group_list, "id"):
            warn("need to pass in a list")
            group_list = [group_list]

        for group in group_list:
            # make sure the group is in the model
            if group.id not in self.groups:
                logger.warning(f"{group!r} not in {self!r}. Ignored.")
            else:
                self.groups.remove(group)
                group._model = None

    def get_associated_groups(
        self, element: Union[Reaction, Gene, Metabolite]
    ) -> List[Group]:
        """Get list of groups for element.

        Returns a list of groups that an element (reaction, metabolite, gene)
        is associated with.

        Parameters
        ----------
        element: `cobra.Reaction`, `cobra.Metabolite`, or `cobra.Gene`

        Returns
        -------
        list of `cobra.Group`
            All groups that the provided object is a member of
        """
        # check whether the element is associated with the model
        return [g for g in self.groups if element in g.members]

    def add_cons_vars(
        self, what: Union[List["CONS_VARS"], Tuple["CONS_VARS"]], **kwargs
    ) -> None:
        """Add constraints and variables to the model's mathematical problem.

        Useful for variables and constraints that can not be expressed with
        reactions and simple lower and upper bounds.

        Additions are reversed upon exit if the model itself is used as
        context.

        Parameters
        ----------
        what : list or tuple of optlang variables or constraints.
           The variables or constraints to add to the model. Must be of
           class `optlang.interface.Variable` or
           `optlang.interface.Constraint`.
        **kwargs : keyword arguments
           Passed to solver.add()
        """
        add_cons_vars_to_problem(self, what, **kwargs)

    def remove_cons_vars(
        self, what: Union[List["CONS_VARS"], Tuple["CONS_VARS"]]
    ) -> None:
        """Remove variables and constraints from problem.

        Remove variables and constraints from the model's mathematical
        problem.

        Remove variables and constraints that were added directly to the
        model's underlying mathematical problem. Removals are reversed
        upon exit if the model itself is used as context.

        Parameters
        ----------
        what : list or tuple of optlang variables or constraints.
           The variables or constraints to add to the model. Must be of
           class `optlang.interface.Variable` or
           `optlang.interface.Constraint`.
        """
        remove_cons_vars_from_problem(self, what)

    @property
    def problem(self) -> "optlang.interface":
        """Get the interface to the model's underlying mathematical problem.

        Solutions to cobra models are obtained by formulating a mathematical
        problem and solving it. Cobrapy uses the optlang package to
        accomplish that and with this property you can get access to the
        problem interface directly.

        Returns
        -------
        optlang.interface
            The problem interface that defines methods for interacting with
            the problem and associated solver directly.
        """
        return self.solver.interface

    @property
    def variables(self) -> "Container":
        """Get the mathematical variables in the cobra model.

        In a cobra model, most variables are reactions. However,
        for specific use cases, it may also be useful to have other types of
        variables. This property defines all variables currently associated
        with the model's problem.

        Returns
        -------
        optlang.container.Container
            A container with all associated variables.
        """
        return self.solver.variables

    @property
    def constraints(self) -> "Container":
        """Get the constraints in the cobra model.

        In a cobra model, most constraints are metabolites and their
        stoichiometries. However, for specific use cases, it may also be
        useful to have other types of constraints. This property defines all
        constraints currently associated with the model's problem.

        Returns
        -------
        optlang.container.Container
            A container with all associated constraints.
        """
        return self.solver.constraints

    @property
    def boundary(self) -> List[Reaction]:
        """Boundary reactions in the model.

        Reactions that either have no substrate or product.

        Returns
        -------
        list
            A list of reactions that either have no substrate or product and
            only one metabolite overall.
        """
        return [rxn for rxn in self.reactions if rxn.boundary]

    @property
    def exchanges(self) -> List[Reaction]:
        """Exchange reactions in model.

        Reactions that exchange mass with the exterior. Uses annotations
        and heuristics to exclude non-exchanges such as sink reactions.

        Returns
        -------
        list
            A list of reactions that satisfy the conditions for exchange reactions.

        See Also
        --------
        cobra.medium.find_boundary_types
        """
        return find_boundary_types(self, "exchange", None)

    @property
    def demands(self) -> List[Reaction]:
        """Demand reactions in model.

        Irreversible reactions that accumulate or consume a metabolite in
        the inside of the model.

        Returns
        -------
        list
            A list of reactions that are demand reactions (reactions that
            accumulate/consume a metabolite irreversibly).

        See Also
        --------
        cobra.medium.find_boundary_types
        """
        return find_boundary_types(self, "demand", None)

    @property
    def sinks(self) -> List[Reaction]:
        """Sink reactions in model.

        Reversible reactions that accumulate or consume a metabolite in
        the inside of the model.

        Returns
        -------
        list
            A list of reactions that are demand reactions (reactions that
            accumulate/consume a metabolite reversibly).

        See Also
        --------
        cobra.medium.find_boundary_types
        """
        return find_boundary_types(self, "sink", None)

    def _populate_solver(
        self,
        reaction_list: List[Reaction],
        metabolite_list: Optional[List[Metabolite]] = None,
    ) -> None:
        """Populate attached solver with constraints and variables.

        Populate attached solver with constraints and variables that
        model the provided reactions.

        Parameters
        ----------
        reaction_list: list
            A list of cobra.Reaction to add to the solver. This list will be
            constrained.
        metabolite_list: list, optional
            A list of cobra.Metabolite  to add to the solver. This list will be
            constrained (default None).
        """
        constraint_terms = AutoVivification()
        to_add = []
        if metabolite_list is not None:
            for met in metabolite_list:
                to_add += [self.problem.Constraint(Zero, name=met.id, lb=0, ub=0)]
        self.add_cons_vars(to_add)

        for reaction in reaction_list:
            if reaction.id not in self.variables:
                forward_variable = self.problem.Variable(reaction.id)
                reverse_variable = self.problem.Variable(reaction.reverse_id)
                self.add_cons_vars([forward_variable, reverse_variable])
            else:
                reaction = self.reactions.get_by_id(reaction.id)
                forward_variable = reaction.forward_variable
                reverse_variable = reaction.reverse_variable
            for metabolite, coeff in reaction.metabolites.items():
                if metabolite.id in self.constraints:
                    constraint = self.constraints[metabolite.id]
                else:
                    constraint = self.problem.Constraint(
                        Zero, name=metabolite.id, lb=0, ub=0
                    )
                    self.add_cons_vars(constraint, sloppy=True)
                constraint_terms[constraint][forward_variable] = coeff
                constraint_terms[constraint][reverse_variable] = -coeff

        self.solver.update()
        for reaction in reaction_list:
            reaction = self.reactions.get_by_id(reaction.id)
            reaction.update_variable_bounds()
        for constraint, terms in constraint_terms.items():
            constraint.set_linear_coefficients(terms)

    def slim_optimize(
        self, error_value: Optional[float] = float("nan"), message: Optional[str] = None
    ) -> float:
        """Optimize model without creating a solution object.

        Creating a full solution object implies fetching shadow prices and
        flux values for all reactions and metabolites from the solver
        object. This necessarily takes some time and in cases where only one
        or two values are of interest, it is recommended to instead use this
        function which does not create a solution object returning only the
        value of the objective. Note however that the `optimize()` function
        uses efficient means to fetch values so if you need fluxes/shadow
        prices for more than say 4 reactions/metabolites, then the total
        speed increase of `slim_optimize` versus `optimize` is  expected to
        be small or even negative depending on how you fetch the values
        after optimization.

        Parameters
        ----------
        error_value : float, None
           The value to return if optimization failed due to e.g.
           infeasibility. If None, raise `OptimizationError` if the
           optimization fails (default float("nan")).
        message : str, optional
           Error message to use if the model optimization did not succeed (default
           None).

        Returns
        -------
        float
            The objective value. Returns the error value if optimization failed and
            error_value was not None.

        Raises
        ------
        OptimizationError
            If error_value was set as None and the optimization fails.
        """
        self.solver.optimize()
        if self.solver.status == optlang.interface.OPTIMAL:
            return self.solver.objective.value
        elif error_value is not None:
            return error_value
        else:
            assert_optimal(self, message)

    def optimize(
        self, objective_sense: Optional[str] = None, raise_error: bool = False
    ) -> "Solution":
        """
        Optimize the model using flux balance analysis.

        Parameters
        ----------
        objective_sense : {None, 'maximize' 'minimize'}, optional
            Whether fluxes should be maximized or minimized. In case of None,
            the previous direction is used (default None).
        raise_error : bool
            If true, raise an OptimizationError if solver status is not
             optimal (default False).

        Returns
        -------
        Solution

        Notes
        -----
        Only the most commonly used parameters are presented here.  Additional
        parameters for cobra.solvers may be available and specified with the
        appropriate keyword argument.

        """
        original_direction = self.objective.direction
        self.objective.direction = {"maximize": "max", "minimize": "min"}.get(
            objective_sense, original_direction
        )
        self.slim_optimize()
        solution = get_solution(self, raise_error=raise_error)
        self.objective.direction = original_direction
        return solution

    def repair(
        self, rebuild_index: bool = True, rebuild_relationships: bool = True
    ) -> None:
        """Update all indexes and pointers in a model.

        Parameters
        ----------
        rebuild_index : bool
            rebuild the indices kept in reactions, metabolites and genes
        rebuild_relationships : bool
             reset all associations between genes, metabolites, model and
             then re-add them.
        """
        if rebuild_index:  # DictList indexes
            self.reactions._generate_index()
            self.metabolites._generate_index()
            self.genes._generate_index()
            self.groups._generate_index()
        if rebuild_relationships:
            for met in self.metabolites:
                met._reaction.clear()
            for gene in self.genes:
                gene._reaction.clear()
            for rxn in self.reactions:
                rxn.update_genes_from_gpr()
                for met in rxn._metabolites:
                    met._reaction.add(rxn)

        # point _model to self
        for dict_list in (self.reactions, self.genes, self.metabolites, self.groups):
            for entity in dict_list:
                entity._model = self

    @property
    def objective(self) -> Union[optlang.Objective]:
        """Get the solver objective.

        With optlang, the objective is not limited to a simple linear summation of
        individual reaction fluxes, making the return value ambiguous.

        Henceforth, use `cobra.util.solver.linear_reaction_coefficients` to
        get a dictionary of reactions with their linear coefficients (empty
        if there are none).

        """
        return self.solver.objective

    @objective.setter
    def objective(
        self,
        value: Union[
            Dict[Reaction, "Container"],
            str,
            int,
            Reaction,
            "optlang.interface.Objective",
            Basic,
        ],
    ) -> None:
        """Set the solver objective.

        Parameters
        ----------
        value: dict or str or int or Reaction or  optlang.interface.Container, Reaction
        or Basic
            The set value can be dictionary (reactions as keys, linear coefficients as
            values), string (reaction identifier), int (reaction index), Reaction or
            problem.Objective or sympy expression directly interpreted as objectives.

        When using in a context, this attribute can be set temporarily.
        """
        if isinstance(value, Basic):
            value = self.problem.Objective(value, sloppy=False)
        if not isinstance(value, (dict, optlang.interface.Objective)):
            try:
                reactions = self.reactions.get_by_any(value)
            except KeyError:
                raise ValueError("invalid objective")
            value = {rxn: 1 for rxn in reactions}
            # TODO - check that it is reset with context.
        set_objective(self, value, additive=False)

    @property
    def objective_direction(self) -> str:
        """Get the objective direction.

        Returns
        -------
        str
            Objective direction as string. Should be "max" or "min".
        """
        return self.solver.objective.direction

    @objective_direction.setter
    @resettable
    def objective_direction(self, value: str) -> None:
        """Set the objective direction.

        When used in a context, this attribute is set temporarily.

        Parameters
        ----------
        value: {"max", "min"}
            String of "max" or "min" for direction.

        Raises
        ------
        ValueError
            If given direction isn't max or min.
        """
        value = value.lower()
        if value.startswith("max"):
            self.solver.objective.direction = "max"
        elif value.startswith("min"):
            self.solver.objective.direction = "min"
        else:
            raise ValueError(f"Unknown objective direction '{value}'.")

    def summary(
        self,
        solution: Optional["Solution"] = None,
        fva: Union["pd.DataFrame", float, None] = None,
    ) -> "ModelSummary":
        """
        Create a summary of the exchange fluxes of the model.

        Parameters
        ----------
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution (default None).
        fva : pd.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched (default None).

        Returns
        -------
        cobra.ModelSummary

        See Also
        --------
        Reaction.summary
        Metabolite.summary

        """
        from cobra.summary import ModelSummary

        return ModelSummary(model=self, solution=solution, fva=fva)

    def __enter__(self) -> "Model":
        """Record future changes to the model.

        Record all future changes to the model, undoing them when a call to
        __exit__ is received. Creates a new context and adds it to the stack.

        Returns
        -------
        cobra.Model
            Returns the model with context added.
        """
        try:
            self._contexts.append(HistoryManager())
        except AttributeError:
            self._contexts = [HistoryManager()]

        return self

    def __exit__(self, type, value, traceback) -> None:
        """Pop the top context manager and trigger the undo functions."""
        context = self._contexts.pop()
        context.reset()

    def merge(
        self,
        right: "Model",
        prefix_existing: Optional[str] = None,
        inplace: bool = True,
        objective: str = "left",
    ) -> "Model":
        """Merge two models to create a model with the reactions from both models.

        Custom constraints and variables from right models are also copied
        to left model, however note that, constraints and variables are
        assumed to be the same if they have the same name.

        Parameters
        ----------
        right : cobra.Model
            The model to add reactions from
        prefix_existing : string or optional
            Prefix the reaction identifier in the right that already exist
            in the left model with this string (default None).
        inplace : bool
            Add reactions from right directly to left model object.
            Otherwise, create a new model leaving the left model untouched.
            When done within the model as context, changes to the models are
            reverted upon exit (default True).
        objective : {"left", "right", "sum"}
            One of "left", "right" or "sum" for setting the objective of the
            resulting model to that of the corresponding model or the sum of
            both (default "left").

        Returns
        -------
        cobra.Model
            The merged model.
        """
        if inplace:
            new_model = self
        else:
            new_model = self.copy()
            new_model.id = f"{self.id}_{right.id}"
        new_reactions = deepcopy(right.reactions)
        if prefix_existing is not None:
            existing = new_reactions.query(lambda rxn: rxn.id in self.reactions)
            for reaction in existing:
                reaction.id = f"{prefix_existing}{reaction.id}"
        new_model.add_reactions(new_reactions)
        interface = new_model.problem
        new_vars = [
            interface.Variable.clone(v)
            for v in right.variables
            if v.name not in new_model.variables
        ]
        new_model.add_cons_vars(new_vars)
        new_cons = [
            interface.Constraint.clone(c, model=new_model.solver)
            for c in right.constraints
            if c.name not in new_model.constraints
        ]
        new_model.add_cons_vars(new_cons, sloppy=True)
        new_model.objective = dict(
            left=self.objective,
            right=right.objective,
            sum=self.objective.expression + right.objective.expression,
        )[objective]
        return new_model

    def _repr_html_(self) -> str:
        """Get HTML represenation of the model.

        Returns
        -------
        str
            Model representation as HTML string.
        """
        return f"""
        <table>
            <tr>
                <td><strong>Name</strong></td>
                <td>{self.id}</td>
            </tr><tr>
                <td><strong>Memory address</strong></td>
                <td>{f"{id(self):x}"}</td>
            </tr><tr>
                <td><strong>Number of metabolites</strong></td>
                <td>{len(self.metabolites)}</td>
            </tr><tr>
                <td><strong>Number of reactions</strong></td>
                <td>{len(self.reactions)}</td>
            </tr><tr>
                <td><strong>Number of genes</strong></td>
                <td>{len(self.genes)}</td>
            </tr><tr>
                <td><strong>Number of groups</strong></td>
                <td>{len(self.groups)}</td>
            </tr><tr>
                <td><strong>Objective expression</strong></td>
                <td>{format_long_string(str(self.objective.expression), 100)}</td>
            </tr><tr>
                <td><strong>Compartments</strong></td>
                <td>{", ".join(v if v else k for k, v in
                               self.compartments.items())}</td>
            </tr>
          </table>"""
